"""
梳理上传过来的DICOM数据为各种ASL数据和其他数据
Version 2.03.04

2022-02-01：1.03.04 => 2.03.04：修改逻辑，T1、T2、DWI、CTP、ASL作为第一层
2021-11-19，增加了接受后延迟标记重复的参数
2021-11-14，替换了原理应用UCLA输出到自己编译的java的JAR包，时间缩短了狠很多
姚泽山
"""

import pydicom
import sys
import os
import shutil
from .AnImageDataSorterSiemens5 import AnImageDataSorterSiemens5
from .AnImageDataSorterSiemens4 import AnImageDataSorterSiemens4
from .AnImageDataSorterSiemensPASL import AnImageDataSorterSiemensPASL
from .AnImageDataSorterGEeASL7 import AnImageDataSorterGE7
from .AnImageDataSorterGEpCASL13 import AnImageDataSorterGE13
from .AnImageDataSorterPhilips13 import AnImageDataSorterPhilips13
from data_sorter.old import AnImageDataSorterMISC
import json


class AnImageDataSorter:

    def __init__(self):

        self.raw_folder = os.path.join(os.getcwd(), '输入', 'Siemens-5delay-raw')
        self.archive_folder = os.path.join(os.getcwd(), '输入', 'Siemens-5delay-archive')
        self.intermediate_folder = os.path.join(os.getcwd(), 'md')
        self.output_path = os.path.join(os.getcwd(), 'AnImage输出')

        self.archive_bottom_folder = ''
        self.data_type = ''
        self.slice_num = 0
        self.raw_data_dict = {}
        self.ge7 = False
        self.asl_type = ''
        self.asl_institution = ''
        self.seriesdelay = 5
        self.delay_time = []
        self.delay_rep = []
        self.label_dur = []
        self.debugMode = False
        self.nonzip = True
        self.series_group_asl = {}
        self.seriesdic = {}
        self.seriesm0 = []
        self.seriesm02 = []
        self.seriescbf = []
        self.seriesadc = []
        self.seriesacbv = []
        self.seriesatt = []
        self.seriesmcbf = []
        self.seriesdwi = []
        self.seriest1 = []
        self.seriest2 = []
        self.seriesatt = []
        self.seriessuv = []
        self.pasl_extra_factor = None
        self.modify = 0

    def set_path(self):
        self.raw_folder = sys.argv[1]
        self.output_path = sys.argv[2]
        try:
            self.archive_bottom_folder = sys.argv[3]
        except:
            self.archive_bottom_folder = ''

    def DebugmodeOn(self):
        self.debugMode = True

    def DebugmodeOff(self):
        self.debugMode = False

    def ZipOff(self):
        self.nonzip = True

    def ZipOn(self):
        self.nonzip = False

    def perform_sorting(self):
        if self.seriessuv:
            AnImageDataSorterMISC.parse_suv_from_archive_folder(self.intermediate_folder, self.seriessuv)

        if self.seriesdwi:
            AnImageDataSorterMISC.parse_dwi_from_archive_folder(self.intermediate_folder, self.seriesdwi,
                                                                self.data_type)

        if self.seriesadc:
            AnImageDataSorterMISC.parse_adc_from_archive_folder(self.intermediate_folder, self.seriesadc)

        if self.seriest1:
            AnImageDataSorterMISC.parse_t1_from_archive_folder(self.intermediate_folder, self.seriest1)

        if self.seriest2:
            AnImageDataSorterMISC.parse_t2_from_archive_folder(self.intermediate_folder, self.seriest2)

        for dict_key in self.series_group_asl.keys():  # 不同序列
            print(dict_key)

            for series_time_key in self.series_group_asl[dict_key].keys():  # 多延迟eASL若多次出现
                if dict_key == "GE_7_eASL":
                    aids = AnImageDataSorterGE7()
                    aids.Sorter(self.intermediate_folder, folder_suffix = str(series_time_key),
                                seriesdic=self.series_group_asl[dict_key][series_time_key],
                                delay_time=self.delay_time, label_dur=self.label_dur)

                if dict_key == "GE_1_pCASL":
                    aids = AnImageDataSorterGE13()
                    aids.Sorter(self.intermediate_folder,
                                series=self.series_group_asl[dict_key][series_time_key])

                if dict_key == "WangJJ_pCASL":
                    aids = AnImageDataSorterSiemens5()
                    aids.Sorter(self.intermediate_folder, self.series_group_asl[dict_key][series_time_key]["RAW"],
                                delay_time=self.delay_time, delay_rep=self.delay_rep)

                if dict_key == "WangJJ_pCASL_old":
                    aids = AnImageDataSorterSiemens4()
                    aids.Sorter(self.intermediate_folder, series_dic=self.series_group_asl[dict_key][series_time_key],
                                delay_time=self.delay_time, delay_rep=self.delay_rep)

                if dict_key == "SI_1_PASL":
                    aids = AnImageDataSorterSiemensPASL()
                    print("启动PASL分析程序")
                    aids.Sorter(self.intermediate_folder, self.series_group_asl[dict_key][series_time_key],
                                delay_time=self.delay_time, delay_rep=self.delay_rep, label_dur=self.label_dur,
                                extra_factor=self.pasl_extra_factor)

                if dict_key == "PH_1_pCASL":
                    aids = AnImageDataSorterPhilips13()
                    aids.Sorter(self.intermediate_folder, series=self.series_group_asl[dict_key][series_time_key])

                # TODO 需要梳理
                # elif self.data_type == 'SIEMENS':
                #     # 若采用了美国的以前的代码的输出，这里会有不同的灌注文件
                #     if series_group[0] == 'UCLA':
                #         aids = AnImageDataSorterSiemens_UCLA()
                #         aids.Sorter(self.intermediate_folder, self.seriesm0, self.seriesdelay,
                #                     init_delay=self.delay_time, seriescbf=self.seriescbf,
                #                     seriesacbv=self.seriesacbv, seriesatt=self.seriesatt, seriesmcbf=self.seriesmcbf)
                #     elif 1 <= series_group[0] <= 3:
                #         aids = AnImageDataSorterSiemens13()
                #         aids.Sorter(self.intermediate_folder, self.seriesm0, self.seriesdelay,
                #                     init_delay=self.delay_time, delay_rep=self.delay_rep)
                #     elif series_group[0] == 'PASL':
                #         aids = AnImageDataSorterSiemensPASL()
                #         aids.Sorter(self.intermediate_folder, self.seriesm0, self.seriesdelay,
                #                     init_delay=self.delay_time, delay_rep=self.delay_rep)
                #

        if not self.nonzip:
            shutil.make_archive(self.output_path, 'zip', self.intermediate_folder)
            self.delete_intermediate()

    def check_zipped(self):
        if self.archive_bottom_folder == 'nonzip':
            self.nonzip = True
        self.create_intermediate()

    def create_intermediate(self):
        if self.nonzip:
            self.intermediate_folder = self.output_path
        os.makedirs(self.intermediate_folder, exist_ok=True)

    def delete_intermediate(self):
        shutil.rmtree(self.intermediate_folder)

    def validate_series(self, input_series_list):
        series_list = input_series_list
        output_series = []
        # 查看每一个series
        for series_idx in series_list:
            detected_asl_type = []
            recognize_flag = False
            target_series = self.raw_data_dict[series_idx]
            temp = self._from_dict_read_dicom_meta(series_idx)
            print('序列名称: ' + temp[0x0008, 0x103e].value)

            # 排除彩色图和图片数少于15的
            try:
                slice_number = temp[0x0019, 0x100A].value
            except:
                slice_number = len(target_series)
            if 'MONOCHROME' not in (temp[(0x0028, 0x0004)].value if (0x0028, 0x0004) in temp else ''):
                print('series is not grayscale, not included ' + temp[0x0008, 0x103e].value)
            elif slice_number < 11:
                print('series has less than 15 slices, not included')

            else:
                # 检查ASL之外的
                if 'T1' in target_series[0].upper() or 'FSPGR' in target_series[0].upper() or \
                        ("AA" in target_series[0].upper() and "SCOUT" in target_series[0].upper()):
                    recognize_flag = True
                    self.seriest1.append(target_series)

                elif 'T2' in target_series[0].upper() and '3' not in target_series[0]:
                    recognize_flag = True
                    self.seriest2.append(target_series)

                elif 'ADC' in target_series[0] and 'eADC' not in target_series[0]:
                    recognize_flag = True
                    detected_asl_type = "ADC"
                    self.seriesadc.append(target_series)

                elif 'Apparent Diffusion Coefficient' in target_series[0] and 'Exponential' not in target_series[0]:
                    recognize_flag = True
                    detected_asl_type = "ADC"
                    self.seriesadc.append(target_series)

                elif 'DWI' in target_series[0] or 'TRACEW' in target_series[0]:
                    recognize_flag = True
                    self.seriesdwi.append(target_series)

                elif [0x0054, 0x0016] in temp and 'FDG' in temp[0x0054, 0x0016][0]:
                    recognize_flag = True
                    self.seriessuv.append(target_series)

                # 检查飞利浦的ASL
                if self.data_type == 'Philips':
                    if 'SOURCE' in target_series[0]:
                        recognize_flag = True
                        self._dictionary_helper_function("PH_1_pCASL", "M0", target_series)

                    elif '3D_pCASL' in target_series[0]:
                        recognize_flag = True
                        self._dictionary_helper_function("PH_1_pCASL", "CBF", target_series)

                # 检查通用电气的ASL
                elif self.data_type == 'GE MEDICAL SYSTEMS':

                    # GE 多延迟
                    if 'eASL' in target_series[0]:

                        try:
                            series_time = int(temp[0x0008, 0x0031].value)
                        except:
                            series_time = 0
                        if series_time not in self.seriesdic.keys():
                            self.seriesdic[series_time] = {}

                        if 'Raw' in target_series[0]:
                            recognize_flag = True
                            detected_asl_type = "eASL_RAW"
                            self._dictionary_helper_function("GE_7_eASL", "RAW", target_series, series_time)

                        elif 'Transit delay' in target_series[0]:
                            recognize_flag = True
                            detected_asl_type = "eASL_ATT"
                            self._dictionary_helper_function("GE_7_eASL", "ATT", target_series, series_time)

                        elif 'Transit corrected' in target_series[0].replace('-', ' '):
                            recognize_flag = True
                            detected_asl_type = "eASL_CBF"
                            self._dictionary_helper_function("GE_7_eASL", "CORR_CBF", target_series, series_time)
                        elif 'CBF' in target_series[0].replace('Flow', 'CBF'):
                            recognize_flag = True
                            detected_asl_type = "eASL_CBF"
                            self._dictionary_helper_function("GE_7_eASL", "STAN_CBF", target_series, series_time)

                    # GE 单延迟
                    else:
                        if 'ASL' in target_series[0] or '3D ASL' in target_series[0]:
                            recognize_flag = True
                            target_series[0] = [target_series[0], temp[0x0018, 0x0082].value]  # 增加后延迟时间的标识
                            self._dictionary_helper_function("GE_1_pCASL", "PWPD", target_series)

                        elif 'CBF' in target_series[0] or 'Cerebral Blood Flow' in target_series[0]:
                            recognize_flag = True
                            detected_asl_type = "3DASL_CBF"
                            target_series[0] = [target_series[0], temp[0x0018, 0x0082].value]  # 增加后延迟时间的标识
                            self._dictionary_helper_function("GE_1_pCASL", "CBF", target_series)

                elif self.data_type == 'SIEMENS':
                    if "parameter_fit" in target_series[0].lower():
                        detected_asl_type = "PASL"
                        recognize_flag = True
                        self._dictionary_helper_function("SI_1_PASL", "PARAMETER_FIT", target_series)

                    elif "pd_tse" in target_series[0].lower() or \
                            ("m0" in target_series[0].lower() and "tgse" not in target_series[0].lower()):
                        detected_asl_type = "PASL"
                        recognize_flag = True
                        self._dictionary_helper_function("SI_1_PASL", "3DPASL_M0", target_series)
                    elif "asl_3d_tra" in target_series[0].lower():
                        detected_asl_type = "PASL"
                        recognize_flag = True
                        self._dictionary_helper_function("SI_1_PASL", "3DPASL_PERF", target_series)
                    elif "asl_2d_tra" in target_series[0].lower():
                        detected_asl_type = "PASL"
                        recognize_flag = True
                        self._dictionary_helper_function("SI_1_PASL", "2DPASL_RAW", target_series)

                    elif "tgse_pcasl" in target_series[0].lower():
                        detected_asl_type = "WangJJ_pCASL"
                        recognize_flag = True
                        self._dictionary_helper_function("WangJJ_pCASL", "RAW", target_series)

                    elif "_te00" in target_series[0].lower():
                        if "_ti4000" in target_series[0].lower():
                            if "ss" in target_series[0].lower():
                                self._dictionary_helper_function("WangJJ_pCASL_old", "M0", target_series)
                                detected_asl_type = "WangJJ_pCASL_old"
                                recognize_flag = True
                        else:
                            delay_time = int(target_series[0].lower().split("ti")[1])
                            if "ns" in target_series[0].lower():
                                target_series[0] = [target_series[0], delay_time]  # 增加后延迟时间的标识
                                self._dictionary_helper_function("WangJJ_pCASL_old", "Control", target_series)
                                detected_asl_type = "WangJJ_pCASL_old"
                                recognize_flag = True
                            elif "ss" in target_series[0].lower():
                                target_series[0] = [target_series[0], delay_time]  # 增加后延迟时间的标识
                                self._dictionary_helper_function("WangJJ_pCASL_old", "Label", target_series)
                                detected_asl_type = "WangJJ_pCASL_old"
                                recognize_flag = True

            # 数据是否被识别
            if not recognize_flag:
                print('文件被排除')

            # 若被识别查看图片数量
            else:
                print("文件识别成功")
                expected_num = [len(self.raw_data_dict[series_idx]) - 1]

                if detected_asl_type == "3DASL_CBF":
                    expected_num = [expected_num[0] * 2]
                elif detected_asl_type == "eASL_CBF":
                    expected_num = [expected_num[0] * 10, expected_num[0] / 7 * 10]
                elif detected_asl_type == "eASL_ATT":
                    expected_num = [expected_num[0] * 10]
                elif detected_asl_type == "eASL_RAW":
                    expected_num = [expected_num[0] / 9 * 10]
                elif detected_asl_type == "ADC":
                    expected_num = [expected_num[0] * 2]

                try:
                    slice_num = temp[0x0020, 0x1002].value
                except:
                    print("slice number not found, assume complete")
                    slice_num = expected_num[0]

                if slice_num not in expected_num:
                    raise Exception('slice missing!')

                output_series.append(target_series[0])

        # TODO 这里需要修改，应该不需要了
        for key in list(self.seriesdic.keys()):
            if not self.seriesdic.get(key):
                del self.seriesdic[key]

        self.seriesdelay = len(self.seriesm0)

        return output_series

        # if self.data_type == 'SIEMENS':
        #     for series_idx in series_list:
        #
        #         if self.raw_data_dict[series_idx][0].lower().startswith('tgse_pcasl') or self.raw_data_dict[series_idx][
        #             0] == 'Parameter_Fit':
        #             self.seriesm0 = self.raw_data_dict[series_idx]
        #
        #         elif self.raw_data_dict[series_idx][0] == 'MoCoSeries':  # 针对友谊医院的2DPASL+M0，头动校正后的数据
        #             self.seriesm0 = self.raw_data_dict[series_idx]
        #
        #         elif 'acbv' in self.raw_data_dict[series_idx][0] and (
        #                 'NonMosaic' in self.raw_data_dict[series_idx][0] or 'asl-' in self.raw_data_dict[series_idx][
        #             0]):
        #             self.seriesacbv = self.raw_data_dict[series_idx]
        #
        #         elif 'att' in self.raw_data_dict[series_idx][0] and (
        #                 'NonMosaic' in self.raw_data_dict[series_idx][0] or 'asl-' in self.raw_data_dict[series_idx][
        #             0]):
        #             self.seriesatt = self.raw_data_dict[series_idx]
        #
        #         elif 'mcbf' in self.raw_data_dict[series_idx][0] and (
        #                 'NonMosaic' in self.raw_data_dict[series_idx][0] or 'asl-' in self.raw_data_dict[series_idx][
        #             0]):
        #             self.seriesmcbf = self.raw_data_dict[series_idx]
        #
        #         elif 'cbf' in self.raw_data_dict[series_idx][0] and (
        #                 'NonMosaic' in self.raw_data_dict[series_idx][0] or 'asl-' in self.raw_data_dict[series_idx][
        #             0]):
        #             self.seriescbf.append(self.raw_data_dict[series_idx])
        #
        #     if len(self.seriesm0) == 27:
        #         self.seriesdelay = 5
        #     elif len(self.seriesm0) == 23 or len(self.seriesm0) == 17:
        #         self.seriesdelay = 1
        #     elif len(self.seriesm0) == 10 and self.seriesm0[0] == 'Parameter_Fit':
        #         self.seriesdelay = 2
        #     elif len(self.seriesm0) == 92 and self.seriesm0[0] == "MoCoSeries":
        #         print("察觉到了友谊医院2DPASL单延迟")
        #         self.seriesdelay = 1
        #     else:
        #         raise Exception("Invalid Number of Siemens Files!")

    def sort_files_and_check_patient(self):
        input_path = self.raw_folder
        file_list = [os.path.join(dp, f) for dp, dn, filenames in os.walk(input_path) for f in filenames]
        patientdate = 0
        patienttime = 0
        PatientInfo = {}
        PatientInfo['PatientName'] = ''
        PatientInfo['PatientBirthDate'] = ''
        PatientInfo['PatientSex'] = ''
        PatientInfo['PatientAge'] = ''
        PatientInfo['PatientWeight'] = ''
        PatientInfo['SeriesDate'] = ''
        PatientInfo['SeriesTime'] = ''
        for file_idx in range(len(file_list)):
            try:
                image_data = pydicom.dcmread(file_list[file_idx], stop_before_pixels=True)
            except:
                continue
            try:
                if patientdate == 0 or patientdate > image_data[0x0008, 0x0021].value:
                    patientdate = image_data[0x0008, 0x0021].value
                    patienttime = image_data[0x0008, 0x0032].value
                elif patientdate == image_data[0x0008, 0x0021].value:
                    if patienttime == 0 or patienttime > image_data[0x0008, 0x0032].value:
                        patienttime = image_data[0x0008, 0x0032].value
            except:
                patientdate = 0
                patienttime = 0
            if not PatientInfo['PatientName']:
                try:
                    PatientInfo['PatientName'] = str(image_data.PatientName)
                except:
                    PatientInfo['PatientName'] = ''
            if not PatientInfo['PatientBirthDate']:
                try:
                    PatientInfo['PatientBirthDate'] = str(image_data.PatientBirthDate)
                except:
                    PatientInfo['PatientBirthDate'] = ''
            if not PatientInfo['PatientSex']:
                try:
                    PatientInfo['PatientSex'] = str(image_data.PatientSex)
                except:
                    PatientInfo['PatientSex'] = ''
            if not PatientInfo['PatientAge']:
                try:
                    PatientInfo['PatientAge'] = str(image_data.PatientAge)
                except:
                    PatientInfo['PatientAge'] = ''
            if not PatientInfo['PatientWeight']:
                try:
                    PatientInfo['PatientWeight'] = str(image_data.PatientWeight)
                except:
                    PatientInfo['PatientWeight'] = ''
            if not PatientInfo['SeriesDate'] and patientdate != 0:
                try:
                    PatientInfo['SeriesDate'] = str(patientdate)
                except:
                    PatientInfo['SeriesDate'] = ''
            if not PatientInfo['SeriesTime'] and patienttime != 0:
                try:
                    PatientInfo['SeriesTime'] = str(patienttime)
                except:
                    PatientInfo['SeriesTime'] = ''

            if self.data_type == '':
                try:
                    self.data_type = image_data[0x0008, 0x0070].value
                except:
                    self.data_type = ''
            else:
                if [0x0008, 0x0070] in image_data and image_data[0x0008, 0x0070].value != self.data_type:
                    raise Exception("Multi-Manufacturer	in One Series!")

            if [0x0020, 0x0013] in image_data and (self.data_type != 'Philips' or (
                    self.data_type == 'Philips' and image_data[0x0020, 0x0100].value == 1)):
                try:
                    series_desc = str(image_data[0x0008, 0x103E].value)
                except:
                    series_desc = 'temp1234567890'
                series_uid = str(image_data[0x0020, 0x000E].value)
                try:
                    instance_num = int(image_data[0x0020, 0x0013].value)
                except:
                    instance_num = 1
                if series_desc != 'temp1234567890':
                    if series_uid not in self.raw_data_dict.keys():
                        self.raw_data_dict[series_uid] = {}
                        self.raw_data_dict[series_uid][int(0)] = series_desc
                    if 'MoCo_' in series_desc or 'asl-' in series_desc:
                        self.raw_data_dict[series_uid][instance_num + 1] = file_list[file_idx]
                    else:
                        self.raw_data_dict[series_uid][instance_num] = file_list[file_idx]

        json_str = json.dumps(PatientInfo)
        with open(os.path.join(self.intermediate_folder, 'patient.json'), 'w') as json_file:
            json_file.write(json_str)

        series_list = list(self.raw_data_dict.keys())
        return series_list

    def _from_dict_read_dicom_meta(self, series_idx):
        try:
            temp = pydicom.dcmread(self.raw_data_dict[series_idx][1],
                                   stop_before_pixels=True)
        except:
            temp = pydicom.dcmread([i for i in self.raw_data_dict[series_idx].values()][1],
                                   stop_before_pixels=True)
        return temp

    def _dictionary_helper_function(self, asl_type, input_string, target_series, series_time=9999):
        if asl_type not in self.series_group_asl.keys():
            self.series_group_asl[asl_type] = {}
        if series_time not in self.series_group_asl[asl_type].keys():
            self.series_group_asl[asl_type][series_time] = {}
        if input_string not in self.series_group_asl[asl_type][series_time].keys():
            self.series_group_asl[asl_type][series_time][input_string] = []
        self.series_group_asl[asl_type][series_time][input_string].append(target_series)


def ImageDataSort(raw_folder, output_path, delay_time, delay_rep, label_dur, pasl_extra_factor, zip=True, debugmode=False, modify=0):
    ImageSorterClass = AnImageDataSorter()
    ImageSorterClass.modify = modify
    ImageSorterClass.nonzip = not zip
    ImageSorterClass.debugMode = debugmode
    ImageSorterClass.raw_folder = raw_folder
    ImageSorterClass.output_path = output_path
    for i in range(len(delay_time)):
        delay_time[i] = delay_time[i]/1000
    ImageSorterClass.delay_time = delay_time
    ImageSorterClass.delay_rep = delay_rep
    for i in range(len(label_dur)):
        label_dur[i] = label_dur[i]/1000
    ImageSorterClass.label_dur = label_dur
    ImageSorterClass.check_zipped()
    ImageSorterClass.pasl_extra_factor = pasl_extra_factor
    all_series_list = ImageSorterClass.sort_files_and_check_patient()
    # print(json.dumps(all_series_list, sort_keys=True, indent=4))
    valid_series_list = ImageSorterClass.validate_series(all_series_list)
    print(json.dumps(valid_series_list, sort_keys=True, indent=4))
    ImageSorterClass.perform_sorting()
    return 0


if __name__ == "__main__":
    raw_folder = r'D:\WorkDir\DataSet\Sorter_test\GE_1pld'
    output_path = r'D:\WorkDir\DataSet\Sorter_test\GE_1pld'
    delay_time = []
    delay_rep = []
    label_dur = []
    ImageDataSort(raw_folder, output_path, delay_time, delay_rep, label_dur, 8)



import os
import pydicom as pd
from . import AnImageDCMUtility as DCMUtil


# adc
def parse_adc_from_archive_folder(intermediate_folder, series):
    os.makedirs(os.path.join(intermediate_folder, 'dwi'), exist_ok=True)
    os.makedirs(os.path.join(intermediate_folder, 'dwi', 'adc'), exist_ok=True)
    keys = sorted(list(series[0].keys()))
    for slice_idx in range(len(series[0]) - 1):
        image_data = pd.dcmread(series[0][keys[slice_idx + 1]])
        DCMUtil.copy_dcm_slice(image_data, intermediate_folder, 'dwi', 'adc', slice_idx,
                               override_series_desc=True)
    DCMUtil.dcm_to_nii(intermediate_folder, 'dwi', 'adc')


# dwi
def parse_dwi_from_archive_folder(intermediate_folder, series, data_type):
    os.makedirs(os.path.join(intermediate_folder, 'dwi'), exist_ok=True)
    nifti_flags = [['b0', False], ['trace1', False], ['trace2', False], ['trace3', False]]
    keys = sorted(list(series[0].keys()))
    for slice_idx in range(len(series[0]) - 1):
        image_data = pd.dcmread(series[0][keys[slice_idx + 1]])
        if data_type == 'GE MEDICAL SYSTEMS':
            b = image_data[0x0043, 0x1039]
            if b[0] == 0:
                series_type = 'b0'
                nifti_flags[0][1] = True
            elif b[0] == 1000:
                series_type = 'trace1'
                nifti_flags[1][1] = True
            else:
                raise Exception("tracew error")
        elif data_type == "Philips":
            b = image_data[0x0008, 0x103E].value.lower()
            if 'b1000' in b:
                series_type = 'trace1'
                nifti_flags[1][1] = True
        elif data_type == 'SIEMENS':
            b = image_data[0x0018, 0x0024].value.lower()
            if 'b10000' in b:
                series_type = 'trace3'
                nifti_flags[3][1] = True
            elif 'b5000' in b:
                series_type = 'trace2'
                nifti_flags[2][1] = True
            elif 'b1000' in b:
                series_type = 'trace1'
                nifti_flags[1][1] = True
            elif 'b0' in b:
                series_type = 'b0'
                nifti_flags[0][1] = True

        os.makedirs(os.path.join(intermediate_folder, 'dwi', series_type), exist_ok=True)
        DCMUtil.copy_dcm_slice(image_data, intermediate_folder, 'dwi', series_type, slice_idx,
                               override_series_desc=True)

    for nifti_idx in range(4):
        if nifti_flags[nifti_idx][1]:
            DCMUtil.dcm_to_nii(intermediate_folder, 'dwi', '{}'.format(nifti_flags[nifti_idx][0]))


def parse_t1_from_archive_folder(intermediate_folder, series):
    os.makedirs(os.path.join(intermediate_folder, 'struct'), exist_ok=True)
    os.makedirs(os.path.join(intermediate_folder, 'struct', 't1'), exist_ok=True)
    keys = sorted(list(series[0].keys()))
    for slice_idx in range(len(series[0]) - 1):
        image_data = pd.dcmread(series[0][keys[slice_idx + 1]])
        DCMUtil.copy_dcm_slice(image_data, intermediate_folder, 'struct', 't1', slice_idx)
    DCMUtil.dcm_to_nii(intermediate_folder, 'struct', 't1')


def parse_t2_from_archive_folder(intermediate_folder, seriest2):
    os.makedirs(os.path.join(intermediate_folder, 'struct'), exist_ok=True)
    for series_id in range(len(seriest2)):
        t2_name = 't2_{:d}'.format(series_id + 1)
        os.makedirs(os.path.join(intermediate_folder, 'struct', t2_name), exist_ok=True)
        for slice_idx in range(len(seriest2[series_id]) - 1):
            image_data = pd.dcmread(seriest2[series_id][slice_idx + 1])
            DCMUtil.copy_dcm_slice(image_data, intermediate_folder, 'struct', t2_name, slice_idx)
        DCMUtil.dcm_to_nii(intermediate_folder, 'struct', t2_name)


# fdg-suv
def parse_suv_from_archive_folder(intermediate_folder, series):
    os.makedirs(os.path.join(intermediate_folder, 'fdg'), exist_ok=True)
    os.makedirs(os.path.join(intermediate_folder, 'fdg', 'suv'), exist_ok=True)
    keys = sorted(list(series[0].keys()))
    for slice_idx in range(len(series[0]) - 1):
        image_data = pd.dcmread(series[0][keys[slice_idx + 1]])
        DCMUtil.copy_dcm_slice(image_data, intermediate_folder, 'fdg', 'suv', slice_idx,
                               override_series_desc=True)




import json
import shutil
import subprocess

from ._dicom import *
from ._dicom_util import write_dicom_series
from .data_sorter_base import DataSorterASLBase


class DataSorterSiemensPCASL_old(DataSorterASLBase):
    def __init__(self, data_type, output_root, delay_time, delay_rep, label_dur, series_list, calc_c_cbf):
        DataSorterASLBase.__init__(self, data_type, output_root, delay_time, delay_rep, label_dur, series_list, calc_c_cbf)
        self.delay_time = delay_time # 需要与识别到的TI做对比，所以先使用ms单位
        self.M0: SeriesModel
        self.Control: List[SeriesModel] = []
        self.Label: List[SeriesModel] = []

        for series in self.series_list:
            series_description = series.SeriesDescription.lower()
            if '_ti4000' in series_description:
                if 'ss' in series_description:
                    self.M0 = series
            elif 'ns' in series_description:
                self.Control.append(series)
            elif 'ss' in series_description:
                self.Label.append(series)

    def validate_series(self):
        if not hasattr(self, 'M0'):
            raise MissingSequenceError('M0[ss_TE00_TI4000]')
        if len(self.Control) != len(self.Label):
            raise ControlNotEqualLabelError(len(self.Control), len(self.Label))

        # 序列名后四位是否为数字
        def parse_ti(desc: str):
            split_ti = desc.split('ti')
            try:
                return int(split_ti[1])
            except:
                return 0
        control_delay_map = { parse_ti(s.SeriesDescription.lower()): s for s in self.Control}
        label_delay_map = { parse_ti(s.SeriesDescription.lower()): s for s in self.Label}

        def split_key_value(d):
            keys = []
            values = []
            for k, v in d:
                keys.append(k)
                values.append(v)
            return keys, values
        control_key, control_value = split_key_value(sorted(control_delay_map.items(), key=lambda x:x[0]))
        label_key, label_value = split_key_value(sorted(label_delay_map.items(), key=lambda x:x[0]))

        if len(control_key) != len(control_key):
            raise SeriesError('序列名称不规范')
        if control_key != label_key:
            raise SeriesError(f'Control{control_key}与Label{label_value}的TI不一致')
        if control_key != self.delay_time:
            logger.warning(f'识别到到TI{control_key}与所指定的delay_time{self.delay_time}不一致；将使用识别到的TI进行计算')
            self.delay_time = control_key
        self.delay_time = [i/1000 for i in self.delay_time]
        # 使用排序后的结果
        self.Control = control_value
        self.Label = label_value

        control_delay_rep = [s.Repetition for s in self.Control]
        label_delay_rep = [s.Repetition for s in self.Label]
        if control_delay_rep != label_delay_rep:
            raise SeriesError(f'Control{control_delay_rep}与Label{label_delay_rep}的重复次数不一致')
        if control_delay_rep != self.delay_rep:
            logger.warning(f'识别到到rep{control_delay_rep}与所指定的delay_rep{self.delay_rep}不一致；将使用识别到的结果')
            self.delay_rep = control_delay_rep


    def Sorter(self):
        # super(DataSorterSiemensPCASL_old, self).Sorter()
        self.validate_series()

        java_temp = os.path.join(self.output_root, 'java_temp')
        try:
            reconstructed_dir = self.generate_from_raw_folder(java_temp)

            # 计算完后，得到的2D增强Dicom文件，将其重整为2D Dicom；重写方向，位置标签
            def lambda_load_dicom(name):
                mat = pydicom.read_file(os.path.join(reconstructed_dir, name+'.dcm')).pixel_array
                mat = mat.swapaxes(0, 2)
                return mat.swapaxes(0, 1)

            cur_series_number = SeriesNumberStartWith.ASL
            file_list = ["m0", "mcbf", "att", "acbv", "cbf1", "cbf2", "cbf3", "cbf4"]
            scaling_list = [1, 0.1, 0.01, 0.001, 0.1, 0.1, 0.1, 0.1]
            for f, s in zip(file_list, scaling_list):
                # 防止溢出
                image = lambda_load_dicom(f.upper())
                write_dicom_series(image, self.M0, f'asl-{f}', cur_series_number,
                                   os.path.join(self.output_root, 'asl', f), slope=s)
                cur_series_number += 1
        except Exception as e:
            logger.error(e)
            raise e
        finally:
            shutil.rmtree(java_temp)

    def generate_from_raw_folder(self, temp_dir):
        # print("层数 = {:d}".format(self.slice_num))
        out_put = os.path.join(temp_dir, "reconstructed")
        os.makedirs(os.path.join(temp_dir, "reordered"), exist_ok=True)
        os.makedirs(os.path.join(temp_dir, "reconstructed"), exist_ok=True)
        self._make_json()

        for i, instance in enumerate(self.M0):
            instance.save_as(os.path.join(temp_dir, "reordered", "temp_m0_dcm_{:04d}.dcm".format(i)))

        number = 0
        for ctrl in self.Control:
            for instance in ctrl:
                instance.save_as(os.path.join(temp_dir, "reordered", "temp_ctrl_dcm_{:04d}.dcm".format(number)))
                number += 1

        number = 0
        for label in self.Label:
            for instance in label:
                instance.save_as(os.path.join(temp_dir, "reordered", "temp_lbl_dcm_{:04d}.dcm".format(number)))
                number += 1

        jar_folder = os.path.join(os.getcwd(), "UCLA_java_jar")

        print(jar_folder)
        response = subprocess.run(
            [
                "java", "-jar",
                "-Djava.library.path={}".format(os.path.join(jar_folder, "lib")),
                os.path.join(jar_folder, "ASL_yao_JAVA-master.jar"),
                os.path.join(temp_dir, "param.json")
            ]
        )
        if response.returncode != 0:
            raise Exception("Java程序运行失败，编号: {:d}".format(response.returncode))

        if os.path.isfile(os.path.join(self.output_root, "java_temp", "reconstructed", "finished.txt")):
            logger.debug("完成用Java的图像重建，开始用Python来调整标签")

        return os.path.join(temp_dir, "reconstructed")

    def _make_json(self):
        delay_num = len(self.delay_time)
        if delay_num == 0:
            raise Exception('delay_time is null')
        if delay_num != len(self.delay_rep):
            raise DelayNotEqualRepetitionError(delay_num, len(self.delay_rep))

        param_json_dict = {
            "data_type": '5delay_mosaic' if self.data_type in [MRI_Type.MRI_TYPE_5Delay_3D_PCASL_SIEMENS,
                                                               MRI_Type.MRI_TYPE_1Delay_3D_PCASL_SIEMENS] else '4delay_3d',
            "slice_order": '',
            "slice_num": self.M0.SliceNumber,
            "init_delay": self.delay_time[0],
            "delay_num": delay_num,
            "interval": self.delay_time[1] - self.delay_time[0] if delay_num >= 2 else 0.0,
            "delay_rep": self.delay_rep,
            "load_path": os.path.join(self.output_root, "java_temp", "reordered"),
            "save_path": os.path.join(self.output_root, "java_temp", "reconstructed")}

        with open(os.path.join(self.output_root, "java_temp", "param.json"), "w") as fp:
            json.dump(param_json_dict, fp, indent=4)
        print("json文件保存")

    @staticmethod
    def _3D_to_mosaic(array: np.ndarray):
        width = height = int(np.ceil(array.shape[2]))
        def get_slice(array: np.ndarray, slice: int):
            # 超出范围用 0值填充
            if array.shape[2] > slice:
                return array[:, :, slice]
            else:
                return np.zeros(array[:, :, 0].shape, dtype=array.dtype)

        row_img_vector = []
        for offset in range(width):
            col_index = tuple([i * width + offset for i in range(height)])
            col_img = np.vstack([get_slice(array, c) for c in col_index])
            row_img_vector.append(col_img)

        return np.hstack(row_img_vector)




from ._dicom import *
from typing import Tuple, Optional
from ._dicom_util import write_dicom_series

class DataSorterBase:
    def __init__(self, data_type, output_root, series_list: List[SeriesModel]):
        self.output_root = output_root
        # 排个序
        self.series_list = sorted(series_list, key=lambda x: x.SeriesNumber)
        self.data_type = data_type
        logger.info(f'识别类型：{data_type}')

    def _validate_series_base(self):
        for i in range(len(self.series_list)):
            self.series_list[i].validate_series()

    def Sorter(self):
        # 清洗前需要基本验证：是否缺图像、是否完整、是否断层...
        self._validate_series_base()

class DataSorterASLBase(DataSorterBase):
    def __init__(self, data_type, output_root, delay_time, delay_rep, label_dur, series_list, calc_c_cbf):
        DataSorterBase.__init__(self, data_type, output_root, series_list)
        self.delay_time = [i / 1000 for i in delay_time]  # 外部单位毫秒， 内部单位秒
        self.delay_rep = delay_rep
        self.label_dur = [i / 1000 for i in label_dur]
        self.calc_c_cbf = calc_c_cbf

        logger.debug(f'[{data_type}] 延迟时间：{delay_time}')
        logger.debug(f'[{data_type}] 标记时长：{label_dur}')
        logger.debug(f'[{data_type}] 重复次数：{delay_rep}')

    def get_PLD_millisecond(self):
        '''
        获取PLD时间, 毫秒单位需要乘以1000
        :return:
        '''
        return [int(i * 1000) for i in self.delay_time]

    def calc_correct_CBF(self, CBF_array: np.ndarray, ref_series: SeriesModel, series_number:int,
                         ATT_array: Optional[np.ndarray] = None):
        '''
        拥有ATT的ASL使用ATT校准CBF，没有ATT的使用多CBF对应体素最大值校准ATT，并输出CBF_PLD图
        :param CBF_array:
        :param ref_series:
        :param series_number:
        :param ATT_array:
        :return:
        '''
        if not self.calc_c_cbf:
            return
        if len(CBF_array.shape) < 4 or CBF_array.shape[3] == 1:
            logger.warning('单PLD无需校正')
            return

        targets = np.array(self.delay_time, str)
        if ATT_array is None:
            # 寻找多个CBF最大的那个
            corr_CBF = np.maximum.reduce(CBF_array, 3)
            write_dicom_series(corr_CBF * 10, ref_series, f'asl-corr-cbf', series_number,
                               os.path.join(self.output_root, 'asl', 'corr-cbf'), slope=0.1)

            # CBF_pld_index = np.argmax(CBF_array, 3)
            # CBF_pld_map = targets[CBF_pld_index]

            # write_dicom_series(CBF_pld_map * 1000, ref_series, f'asl-corr-att', series_number,
            #                    os.path.join(self.output_root, 'asl', 'corr-att'), dtype=np.float32, slope=0.001)

        else:
            is_back = True
            def back_function(x):
                for idx, num in enumerate(self.delay_time):
                    if num >= x or (x - num) < 0.01:
                        return idx
                return len(self.delay_time) - 1
            def near_function(x):
                abs_diff = np.abs(targets - x)
                return int(np.argmin(abs_diff))


            # 使用 np.vectorize() 创建向量化函数
            vectorized_function = np.vectorize(back_function if is_back else near_function)
            CBF_pld_index = vectorized_function(ATT_array)

            corr_CBF = np.zeros(CBF_pld_index.shape, dtype=np.float32)
            for i in range(len(self.delay_time)):
                corr_CBF[CBF_pld_index == i] = CBF_array[...,i][CBF_pld_index == i]

            write_dicom_series(corr_CBF * 10, ref_series, f'asl-corr-cbf', series_number,
                               os.path.join(self.output_root, 'asl', 'corr-cbf'), slope=0.1)

            # 使用索引获取最接近的目标数组元素的值
            # CBF_pld_map = targets[CBF_pld_index]
            # write_dicom_series(CBF_pld_map * 1000, ref_series, f'asl-corr-att', series_number,
            #                    os.path.join(self.output_root, 'asl', 'corr-att'), dtype=np.float32, slope=0.001)




import os

# import gdcm
import numpy as np
import pydicom
from log import logger
from nibabel.nicom import csareader
from pydicom.errors import InvalidDicomError
from typing import Dict, List

from .errors import *
from ._type import *


def demosaic(mosaic_mat: np.ndarray, slice_num: int):
    n_rows = n_cols = int(np.ceil(np.sqrt(slice_num)))
    img_width, img_height = mosaic_mat.shape[1], mosaic_mat.shape[0]
    img_width //= n_cols
    img_height //= n_rows

    _3d_mat = np.zeros((img_height, img_width, slice_num))
    for i in range(slice_num):
        x_start = i % n_cols * img_width
        y_start = i // n_cols * img_height
        _3d_mat[::,:,i] = mosaic_mat[y_start:y_start+img_height, x_start:x_start+img_width]

    return _3d_mat


def csa_ascii_read_key(csa_ascii_list:list, key:str, default=None):
    value = ''
    exist_flag = False
    for ascii in csa_ascii_list:
        if ascii.startswith(key):
            exist_flag = True
            value = ascii.split('=')[-1].strip().strip('')
    if exist_flag:
        return value
    elif default is None:
        raise CsaHeaderMissingKeyError(key)
    else:
        return default

def get_csa_ascii_list(dataset: pydicom.Dataset) -> list:
    try:
        csa_header_series = csareader.read(DicomRequiredTags.get_key((0x0029, 0x1020), dataset))
        MrPhoenixProtocol = csa_header_series['tags']['MrPhoenixProtocol']['items'][0]
        csa_ascii_begin = '### ASCCONV BEGIN'
        csa_ascii_end = '### ASCCONV END'
        csa_ascii = MrPhoenixProtocol[MrPhoenixProtocol.find(csa_ascii_begin):MrPhoenixProtocol.find(csa_ascii_end)].split('\n')
        return csa_ascii
    except:
        return []

# dicom必要的标签信息
class DicomRequiredTags:
    def __init__(self, dataset: pydicom.Dataset):
        # Patient信息；允许没有
        self.PatientName = self.get_key((0x0010, 0x0010), dataset, '')
        self.PatientID = self.get_key((0x0010, 0x0020), dataset, '')
        self.PatientBirthDate = self.get_key((0x0010, 0x0030), dataset, '')
        self.PatientSex = self.get_key((0x0010, 0x0040), dataset, '')
        self.PatientAge = self.get_key((0x0010, 0x1010), dataset, '')
        self.PatientSize = self.get_key((0x0010, 0x1020), dataset, '')
        self.PatientWeight = self.get_key((0x0010, 0x103D), dataset, '')
        self.PatientPosition = self.get_key((0x0018, 0x5100), dataset, '')

        # Study信息，处理UID其余运行没有
        self.StudyInstanceUID = self.get_key((0x0020, 0x000D), dataset)
        self.StudyDescription = self.get_key((0x0008, 0x1030), dataset, '')
        self.StudyDate = self.get_key((0x0008, 0x0020), dataset, '')
        self.StudyTime = self.get_key((0x0008, 0x0030), dataset, '')
        self.StudyID = self.get_key((0x0020, 0x0010), dataset, '')

        # Series信息
        self.SeriesInstanceUID = self.get_key((0x0020, 0x000E), dataset)
        self.SeriesDescription = self.get_key((0x0008, 0x103E), dataset)
        self.SeriesDate = self.get_key((0x0008, 0x0021), dataset, '')
        self.SeriesTime = self.get_key((0x0008, 0x0031), dataset, '')
        self.SeriesNumber = self.get_key((0x0020, 0x0011), dataset, '')

        # 设备信息
        self.Manufacturer = MANUFACTURER.str2MANUFACTURER(self.get_key((0x0008, 0x0070), dataset))
        self.ManufacturerModelName = self.get_key((0x0008, 0x1090), dataset, UnKnown)
        self.Modality = MODALITY.str2MODALITY(self.get_key((0x0008, 0x0060), dataset))
        self.InstitutionName = self.get_key((0x0008, 0x0080), dataset, '')
        self.ProtocolName = self.get_key((0x0018, 0x1030), dataset, '')

        # 序列名
        self.PulseSequenceName = self.get_key((0x0018, 0x9005), dataset, '')
        self.SamplesPerPixel = self.get_key((0x0028, 0x0002), dataset, 1)
        if self.SamplesPerPixel != 1:
            # 不是原始的灰度图像忽略即可
            raise PhotometricRGBError()

        # 图像信息
        self.SOPInstanceUID = self.get_key((0x0008, 0x0018), dataset)
        self.InstanceNumber = self.get_key((0x0020, 0x0013), dataset)
        self.ImageMat = dataset.pixel_array

        # RealWorldValueMappingSequence = self.get_key((0x0040, 0x9096), dataset, [])
        # # 0028,1053和0028,1052定义的重新缩放斜率和截距，除了飞利浦以外，这是定义 缩放斜率和截距唯一的地方。
        # if len(RealWorldValueMappingSequence) > 0 and self.Manufacturer == MANUFACTURER.kMANUFACTURER_PHILIPS:
        #     self.RescaleSlope = self.get_key((0x0040, 0x9225), RealWorldValueMappingSequence[0], 1)
        #     self.RescaleIntercept = self.get_key((0x0040, 0x9224), RealWorldValueMappingSequence[0], 0)
        # else:
        if True:
            self.RescaleSlope = self.get_key((0x0028, 0x1053), dataset, 1)
            self.RescaleIntercept = self.get_key((0x0028, 0x1052), dataset, 0)

        if len(self.ImageMat.shape) == 3:
            self.dicom_image_type = DICOM_IMAGE_TYPE.DICOM_IMAGE_TYPE_3D
            # 每个切片放在第三维
            self.ImageMat = self.ImageMat.swapaxes(0, 2)
            self.ImageMat = self.ImageMat.swapaxes(0, 1)

            # 3D格式的图像信息在 5200|9229 共享信息 或者 5200|9230 每帧信息中
            # 5200|9229 不存在的话到 5200|9230中找
            try:
                SharedFunctionalGroupsSequence = self.get_key((0x5200, 0x9229), dataset)
                PixelMeasures = self.get_key((0x0028, 0x9110), SharedFunctionalGroupsSequence[0])
                PixelValueTransformation = self.get_key((0x0028, 0x9145), SharedFunctionalGroupsSequence[0])
            except:
                SharedFunctionalGroupsSequence = self.get_key((0x5200, 0x9230), dataset)
                PixelMeasures = self.get_key((0x0028, 0x9110), SharedFunctionalGroupsSequence[0])
                PixelValueTransformation = self.get_key((0x0028, 0x9145), SharedFunctionalGroupsSequence[0])

            self.RescaleSlope = self.get_key((0x0028, 0x1053), PixelValueTransformation[0])
            self.RescaleIntercept = self.get_key((0x0028, 0x1052), PixelValueTransformation[0])
            self.SliceThickness = self.get_key((0x0018, 0x0050), PixelMeasures[0])
            self.SpacingBetweenSlices = self.get_key((0x0018, 0x0088), PixelMeasures[0], self.SliceThickness)
            self.PixelSpacing = self.get_key((0x0028, 0x0030), PixelMeasures[0])

            self.ImagePositionPatientList = []
            Orientation_list = set()
            self.PerFrameFunctionalGroupsSequence = self.get_key((0x5200, 0x9230), dataset)
            for frame in self.PerFrameFunctionalGroupsSequence:
                PlanePositionSequence = self.get_key((0x0020, 0x9113), frame)
                ImagePosition = self.get_key((0x0020, 0x0032), PlanePositionSequence[0])
                PlaneOrientationSequence = self.get_key((0x0020, 0x9116), frame)
                ImageOrientation = self.get_key((0x0020, 0x0037), PlaneOrientationSequence[0])
                self.ImagePositionPatientList.append(ImagePosition)
                Orientation_list.add(tuple(ImageOrientation))

            # todo：有点问题，只有方向无偏角的时候才正确。
            self.SliceLocation = self.ImagePositionPatientList[0][2]

            if len(Orientation_list) != 1:
                raise Exception('3D图像方向不一致')
            self.ImageOrientationPatient = list(list(Orientation_list)[0])


        elif len(self.ImageMat.shape) == 2:
            # 有可能是Mosaic图像，也有可能是普通2D图像
            self.dicom_image_type = DICOM_IMAGE_TYPE.DICOM_IMAGE_TYPE_2D
            self.PixelSpacing = self.get_key((0x0028, 0x0030), dataset)
            self.ImagePositionPatient = self.get_key((0x0020, 0x0032), dataset)
            self.ImageOrientationPatient = self.get_key((0x20, 0x0037), dataset)
            self.SliceLocation = self.get_key((0x0020, 0x1041), dataset)

            # 层厚/像素间距拥有其中一个标签即可
            self.SliceThickness = self.get_key((0x0018, 0x0050), dataset)
            # 层间距不存在则认为层间距=层厚
            self.SpacingBetweenSlices = self.get_key((0x0018, 0x0088), dataset, self.SliceThickness)
        else:
            raise UnSupportDataTypeError

        self.BitsAllocated = self.get_key((0x0028, 0x0100), dataset)
        self.BitsStored = self.get_key((0x0028, 0x0101), dataset)
        self.HighBit = self.get_key((0x0028, 0x0102), dataset)
        self.Rows = self.get_key((0x0028, 0x0010), dataset)
        self.Columns = self.get_key((0x0028, 0x0011), dataset)
        self.PixelRepresentation = self.get_key((0x0028, 0x0103), dataset)
        self.PhotometricInterpretation = self.get_key((0x0028, 0x0004), dataset, 'MONOCHROME2') # 默认最大值白色
        # 什么剖面图
        self.SliceDirection = affine2Orientation(self.ImageOrientationPatient)
        self._raw_dataset = dataset


    @staticmethod
    def get_key(key, dataset: pydicom.Dataset, default=None):
        '''
        key不存在并且default未设置，则抛异常
        :param key:
        :param dataset:
        :param default:
        :return:
        '''
        if key not in dataset:
            if default is None:
                raise DicomMissingTagError(key)
            return default
        else:
            return dataset[key].value

    def get_tag(self, key, default=None):
        return self.get_key(key, self._raw_dataset, default)


class InstanceModel(DicomRequiredTags):
    def __init__(self, dataset: pydicom.Dataset, dcm_path: str):
        super().__init__(dataset)
        self.dcm_path = dcm_path

    def save_as(self, file_name):
        self._raw_dataset.save_as(file_name)


class SeriesModel(List[InstanceModel]):
    def __init__(self, instance:InstanceModel, raw_dataset: pydicom.Dataset):
        list.__init__(self)
        self._unique_uid = set()  # 记录instanceUID；避免重复文件
        self.PatientName = instance.PatientName
        self.PatientID = instance.PatientID
        self.PatientBirthDate = instance.PatientBirthDate
        self.PatientSex = instance.PatientSex
        self.PatientAge = instance.PatientAge
        self.PatientSize = instance.PatientSize
        self.PatientWeight = instance.PatientWeight
        self.PatientPosition = instance.PatientPosition
        self.StudyInstanceUID = instance.StudyInstanceUID
        self.StudyDescription = instance.StudyDescription
        self.StudyDate = instance.StudyDate
        self.StudyTime = instance.StudyTime
        self.StudyID = instance.StudyID
        self.SeriesInstanceUID = instance.SeriesInstanceUID
        self.SeriesDescription = instance.SeriesDescription
        self.SeriesDate = instance.SeriesDate
        self.SeriesTime = instance.SeriesTime
        self.SeriesNumber = instance.SeriesNumber
        self.Manufacturer = instance.Manufacturer
        self.ManufacturerModelName = instance.ManufacturerModelName
        self.Modality = instance.Modality
        self.InstitutionName = instance.InstitutionName
        self.ProtocolName = instance.ProtocolName
        self.SamplesPerPixel = instance.SamplesPerPixel
        self.RescaleSlope = instance.RescaleSlope
        self.RescaleIntercept = instance.RescaleIntercept
        self.BitsAllocated = instance.BitsAllocated
        self.BitsStored = instance.BitsStored
        self.HighBit = instance.HighBit
        self.Rows = instance.Rows
        self.Columns = instance.Columns
        self.SliceThickness = instance.SliceThickness
        self.SpacingBetweenSlices = instance.SpacingBetweenSlices
        self.PixelSpacing = instance.PixelSpacing
        self.PixelRepresentation = instance.PixelRepresentation
        self.PhotometricInterpretation = instance.PhotometricInterpretation

        self.SliceDirection = instance.SliceDirection
        self.PulseSequenceName = instance.PulseSequenceName
        self.SliceNumber = 0

        self.Repetition = 1   # 表示有多少组数据，（拥有控制像, 标记像）重复次数为1/2
        self.dicom_image_type = instance.dicom_image_type
        if self.dicom_image_type == DICOM_IMAGE_TYPE.DICOM_IMAGE_TYPE_3D:
            self.PerFrameFunctionalGroupsSequence = instance.PerFrameFunctionalGroupsSequence

        # 序列完整之后再获取坐标信息
        self.ImagePositionPatientList:List
        self.ImagePositionPatient:List
        self.ImageOrientationPatient:List
        self.SliceLocation:List
        self.z_reversed:bool = False    # 默认Z轴为增加

        if self.Modality != MODALITY.MODALITY_MR:
            raise UnSupportDataTypeError(self.Modality)

        if self.Manufacturer == MANUFACTURER.kMANUFACTURER_SIEMENS:
            # 判断是否为MOSAIC
            self.MosaicSliceNumber = instance.get_tag((0x0019, 0x100A), 0)
            if self.MosaicSliceNumber > 0 and self.dicom_image_type == DICOM_IMAGE_TYPE.DICOM_IMAGE_TYPE_2D:
                self.dicom_image_type = DICOM_IMAGE_TYPE.DICOM_IMAGE_TYPE_MOSAIC

                # MOSAIC图像需要进一步解析 CSA头部信息
                csa_header_image = csareader.read(instance.get_tag((0x0029, 0x1010), raw_dataset))
                # 用于辨别方向
                self.SliceNormalVector = csa_header_image['tags']['SliceNormalVector']['items']

            # 读取CSA Header 用于辨别PCASL 或者是 PASL
            csa_ascii_list = get_csa_ascii_list(raw_dataset)
            if self.PulseSequenceName == '':
                self.PulseSequenceName = csa_ascii_read_key(csa_ascii_list, 'tSequenceFileName', '')
            self.asl_type, self.data_type = self.inference_data_type(
                self.Manufacturer, self.PulseSequenceName, self.SeriesDescription)
        elif self.Manufacturer == MANUFACTURER.kMANUFACTURER_PHILIPS:
            self.asl_type, self.data_type = self.inference_data_type(self.Manufacturer,
                                                                             series_description=self.SeriesDescription)
        elif self.Manufacturer == MANUFACTURER.kMANUFACTURER_GE:
            self.asl_type, self.data_type = self.inference_data_type(self.Manufacturer,
                                                                             series_description=self.SeriesDescription)
        elif self.Manufacturer == MANUFACTURER.kMANUFACTURER_UIH:
            self.asl_type, self.data_type = self.inference_data_type(self.Manufacturer,
                                                                             series_description=self.SeriesDescription)
        else:
            # 暂未开发完成
            raise UnSupportDataTypeError('暂不支持除西门子外的数据处理')

    def get_tag(self, key, default=None):
        return self[0].get_tag(key, default)

    def append(self, instance: InstanceModel):
        if instance.SeriesInstanceUID != self.SeriesInstanceUID:
            logger.warning(f'The UID({instance.SeriesInstanceUID}) of the instance is different from that of the series({self.SeriesInstanceUID})')
            return
        # 已记录的文件排除
        if instance.SOPInstanceUID in self._unique_uid:
            logger.debug(f'{instance.SOPInstanceUID} dicom文件重复，被忽略')
            return
        self._unique_uid.add(instance.SOPInstanceUID)
        super().append(instance)

    def finish_handle(self):
        # 所有图像都获取之后、完整后的操作
        # 首先根据instanceNumber进行图片排序
        self.sort(key=lambda x: x.InstanceNumber)
        self.SliceLocation = self[0].SliceLocation
        self.ImageOrientationPatient = self[0].ImageOrientationPatient

        if self.dicom_image_type == DICOM_IMAGE_TYPE.DICOM_IMAGE_TYPE_MOSAIC:
            self.ImagePositionPatient = self[0].ImagePositionPatient

            self.SliceNumber = self.MosaicSliceNumber
            self.Repetition = len(self)  # 不除以2了

            # 判断Z轴正方向还是负方向
            slice_norm_mat = np.zeros((3, 3), dtype=float)
            slice_norm_mat[:, 0] = self.ImageOrientationPatient[:3]
            slice_norm_mat[:, 1] = self.ImageOrientationPatient[3:]
            slice_norm_mat[:, 2] = self.SliceNormalVector
            # 法向量相反，Z序需要翻转
            if np.linalg.det(slice_norm_mat) < 0:
                self.z_reversed = True  # Z轴递减


        elif self.dicom_image_type == DICOM_IMAGE_TYPE.DICOM_IMAGE_TYPE_2D:
            self.ImagePositionPatient = self[0].ImagePositionPatient
            # 一个序列可能包含多次重复扫描，根据SliceLocation分成多次扫描
            # dicom文件由两种排列方式
            # 第一种是每个完整的重复次数放在一起
            # 第二种是相同的层放在一起
            if len(self) < 2:
                return
            first_slice_location = self[0].SliceLocation
            second_slice_location = self[1].SliceLocation
            self.is_second_arrangement_mode = (first_slice_location == second_slice_location)
            # 是否为第二种排列方式, 第二种模式表示同一位置切片放在一起。
            # 第一种模式表示整个Volume切片放在一起

            if self.is_second_arrangement_mode:
                for i in range(1, len(self)):
                    if self[i].SliceLocation != first_slice_location:
                        self.Repetition = i
                        break
                # 未找到不同的层厚信息则层厚为1
                self.Repetition = 1 if self.Repetition == 0 else self.Repetition
                self.SliceNumber = len(self) / self.Repetition
            else:
                for i in range(1, len(self)):
                    if self[i].SliceLocation == first_slice_location:
                        self.SliceNumber = i
                        break
                # 未发现相同的，层数量等于图片数量
                self.SliceNumber = len(self) if self.SliceNumber == 0 else self.SliceNumber
                self.Repetition = len(self) / self.SliceNumber

            self.SliceNumber = int(self.SliceNumber)
            self.Repetition = int(self.Repetition)

            if self.is_second_arrangement_mode:
                second_slice_position = self[self.Repetition].ImagePositionPatient
            else:
                second_slice_position = self[1].ImagePositionPatient

            # 判断Z轴正方向还是负方向
            self.iSL, self.z_reversed = self.verify_slice_dir(second_slice_position)

        elif self.dicom_image_type == DICOM_IMAGE_TYPE.DICOM_IMAGE_TYPE_3D:
            self.ImagePositionPatientList = self[0].ImagePositionPatientList
            self.ImagePositionPatient = self[0].ImagePositionPatientList[0]

            # 查看图片相同定位的数量
            same_position = {}
            for i in self.ImagePositionPatientList:
                multiValue2tuple = tuple(i)
                if multiValue2tuple not in same_position:
                    same_position[multiValue2tuple] = 0
                same_position[multiValue2tuple] += 1

            self.SliceNumber = len(same_position)
            self.Repetition = list(same_position.values())[0]
            # 多个3D dicom文件
            self.Repetition *= len(self)

            first_slice_location = self.ImagePositionPatientList[0]
            second_slice_location = self.ImagePositionPatientList[1]
            self.is_second_arrangement_mode = (first_slice_location == second_slice_location)
            self.is_third_arrangement_mode = False
            # 有可能出现第三种情况：第三种模式表示 第二模式基础上将label与control像分开放。
            # 第二种模式表示同一位置（SliceLocation）切片放在一起。
            # 第一种模式表示整个Volume切片放在一起
            if self.is_second_arrangement_mode:
                if self.ImagePositionPatientList[self.Repetition // 2] != second_slice_location:
                    self.is_third_arrangement_mode = True
                    self.is_second_arrangement_mode = False

            self.ImagePositionPatientList = [[float(j) for j in i]  for i in same_position]
            # 判断Z轴正方向还是负方向
            second_slice_location = list(same_position)[1]
            self.iSL, self.z_reversed = self.verify_slice_dir(second_slice_location)

        if self.z_reversed:
            logger.info(f'{self.SeriesDescription} reversed')

    def verify_slice_dir(self, second_slice_position):
        # return slice direction: 1 = sag, 2 = coronal, 3 = axial
        iSL = self.SliceDirection.iSL()

        second_slice_pos = second_slice_position[iSL - 1]
        m = self._calc_2d_affine()
        # 通过矩阵计算一下正方向的location与第二张图的location是否同向
        # 同向说明不需要反转，否则需要反转
        x = (0, 0, self.SliceNumber - 1, 1)

        posv = np.matmul(m, x)
        pos = posv[iSL-1]
        return iSL, (second_slice_pos > m[iSL-1][3]) != (pos > m[iSL-1][3])

    def inference_data_type(self, manufacturer: MANUFACTURER, sequence_name:str=None, series_description: str=None)->(ASL_Type,MRI_Type):
        '''
        根据 sequence 或者 series_description 推断数据类型
        :param manufacturer:
        :param sequence_name:
        :param series_description:
        :return:
        '''
        if sequence_name is not None:
            sequence_name = sequence_name.lower()
        if series_description is not None:
            series_description = series_description.lower()
        ASL_type = ASL_Type.ASL_UnKnown
        data_type = MRI_Type.MRI_TYPE_UnKnown
        # 排除一些像截图这类的文件
        if series_description.find('[') > -1 and series_description.find(']') > -1:
            return ASL_type, data_type
        if manufacturer == MANUFACTURER.kMANUFACTURER_SIEMENS:
            if sequence_name is not None and sequence_name != '':
                if self.PulseSequenceName.find('tgse_pcasl') > -1:
                    ASL_type = ASL_Type.ASL_PCASL

                elif self.PulseSequenceName.find('tgse_pasl') > -1:
                    ASL_type, data_type = ASL_Type.ASL_PASL, MRI_Type.MRI_TYPE_3D_PASL_SIEMENS

            if series_description is not None and series_description != '':
                if data_type != MRI_Type.MRI_TYPE_UnKnown:
                    pass
                else:
                    if (series_description.startswith('tgse_pcasl_pld') and
                        not series_description.startswith('tgse_pcasl_pld5')) or \
                            series_description.startswith('tgse_pcasl_m0'):
                        ASL_type, data_type = ASL_Type.ASL_PCASL, MRI_Type.MRI_TYPE_1Delay_3D_PCASL_SIEMENS
                    elif series_description.startswith('tgse_pcasl'):
                        ASL_type, data_type = ASL_Type.ASL_PCASL, MRI_Type.MRI_TYPE_5Delay_3D_PCASL_SIEMENS
                    elif series_description.find('te00_ti') > -1:
                        ASL_type, data_type = ASL_Type.ASL_PCASL, MRI_Type.MRI_TYPE_4Delay_3D_PCASL_SIEMENS_Old
                    elif series_description.startswith('asl_3d') or series_description.find('m0') > -1:
                        ASL_type, data_type = ASL_type.ASL_PASL, MRI_Type.MRI_TYPE_3D_PASL_SIEMENS
                    elif series_description.find('t1') > -1:
                        data_type = MRI_Type.MRI_TYPE_T1
                        # t1 放在t2 flair前面；避免t1 flair被认为是 t2 flair
                    elif series_description.find('flair') > -1 or series_description.find('fluid') > -1:
                        data_type = MRI_Type.MRI_TYPE_T2_FLAIR
                    elif series_description.find('t2') > -1:
                        data_type = MRI_Type.MRI_TYPE_T2
                    elif series_description.find('dwi') > -1 or series_description.find('adc') > -1 or \
                            series_description.find('apparent diffusion coefficient') > -1 or \
                            series_description.find('tracew') > -1:
                        data_type = MRI_Type.MRI_TYPE_DWI

        elif manufacturer == MANUFACTURER.kMANUFACTURER_PHILIPS:
            if sequence_name is not None and sequence_name != '':
                if sequence_name == 'grase':
                    ASL_type, data_type =  ASL_Type.ASL_PCASL, MRI_Type.MRI_TYPE_3D_PCASL_PHILIPS
            if series_description is not None and series_description != '':
                if series_description.find('3d') > -1 and series_description.find('pcasl') > -1:
                    ASL_type, data_type = ASL_Type.ASL_PCASL, MRI_Type.MRI_TYPE_3D_PCASL_PHILIPS
                elif series_description.find('t1') > -1:
                    data_type = MRI_Type.MRI_TYPE_T1
                    # t1 放在t2 flair前面；避免t1 flair被认为是 t2 flair
                elif series_description.find('flair') > -1 or series_description.find('fluid') > -1:
                    data_type = MRI_Type.MRI_TYPE_T2_FLAIR
                elif series_description.find('t2') > -1:
                    data_type = MRI_Type.MRI_TYPE_T2
                elif series_description.find('dwi') > -1 or series_description.find('adc') > -1 or \
                    series_description.find('apparent diffusion coefficient') > -1:
                    data_type = MRI_Type.MRI_TYPE_DWI

        elif manufacturer == MANUFACTURER.kMANUFACTURER_GE:
            if series_description is not None and series_description != '':
                if series_description.find('easl') > -1:
                    ASL_type, data_type = ASL_Type.ASL_PCASL, MRI_Type.MRI_TYPE_eASL_PCASL_GE
                elif (series_description.find('3d') > -1 and series_description.find('asl') > -1) or \
                        series_description.find('cerebral blood flow') > -1 or series_description.find('cbf') > -1:
                    ASL_type, data_type = ASL_Type.ASL_PCASL, MRI_Type.MRI_TYPE_3D_PCASL_GE
                elif series_description.find('t1') > -1:
                    data_type = MRI_Type.MRI_TYPE_T1
                    # t1 放在t2 flair前面；避免t1 flair被认为是 t2 flair
                elif series_description.find('flair') > -1 or series_description.find('fluid') > -1:
                    data_type = MRI_Type.MRI_TYPE_T2_FLAIR
                elif series_description.find('t2') > -1:
                    data_type = MRI_Type.MRI_TYPE_T2
                elif series_description.find('dwi') > -1 or series_description.find('adc') > -1 or \
                        series_description.find('apparent diffusion coefficient') > -1:
                    data_type = MRI_Type.MRI_TYPE_DWI

        elif manufacturer == MANUFACTURER.kMANUFACTURER_UIH:
            if series_description is not None and series_description != '':
                if series_description.startswith('asl_3d'):
                    ASL_type, data_type = ASL_type, MRI_Type.MRI_TYPE_3D_PASL_UIH
                elif series_description.find('t1') > -1:
                    data_type = MRI_Type.MRI_TYPE_T1
                    # t1 放在t2 flair前面；避免t1 flair被认为是 t2 flair
                elif series_description.find('flair') > -1 or series_description.find('fluid') > -1:
                    data_type = MRI_Type.MRI_TYPE_T2_FLAIR
                elif series_description.find('t2') > -1:
                    data_type = MRI_Type.MRI_TYPE_T2

        return ASL_type, data_type

    def validate_series(self):
        '''
        验证数据的完整性以及正确性；计算出层厚、层数量、以及重复次数是否符合要求
        :return:
        '''
        # 检查是否缺少Slice
        if self.dicom_image_type in (DICOM_IMAGE_TYPE.DICOM_IMAGE_TYPE_3D, DICOM_IMAGE_TYPE.DICOM_IMAGE_TYPE_MOSAIC):
            for i, image in enumerate(self):
                if (image.InstanceNumber != self.SliceNumber * i + 1) and (image.InstanceNumber != i + 1):
                    raise MissingSliceError(i * self.SliceNumber + 1)

        elif self.dicom_image_type == DICOM_IMAGE_TYPE.DICOM_IMAGE_TYPE_2D:
            for r in range(self.Repetition):
                first_number = self[r * self.SliceNumber].InstanceNumber
                # 增量：为应对某些number是等差增长，而非+1。但前提是第一张和第二章是不缺损的,才能保证正确
                d = self[r * self.SliceNumber + 1].InstanceNumber - self[r * self.SliceNumber].InstanceNumber
                # 允许每个Volume不连续
                for n in range(self.SliceNumber):
                    no = r * self.SliceNumber + n
                    if self[no].InstanceNumber != first_number + d * n:
                        logger.info(f'未验证通过序列：{self.SeriesDescription}；层数：{self.SliceNumber}，重复次数：{self.Repetition}，'
                                    f'图像类型：{self.dicom_image_type.value}，剖面：{self.SliceDirection}')
                        raise MissingSliceError(first_number + d * n)

        logger.info(f'验证通过序列：{self.SeriesDescription}；层数：{self.SliceNumber}，重复次数：{self.Repetition}，'
                    f'图像类型：{self.dicom_image_type.value}，剖面：{self.SliceDirection}')

    def load(self, rescale=1):
        '''
        加载数序列的图像，4维图像：包含重复次数
        :return:
        '''
        image = None
        if self.dicom_image_type == DICOM_IMAGE_TYPE.DICOM_IMAGE_TYPE_MOSAIC:
            for idx, instance in enumerate(self):
                _3d_image = demosaic(instance.ImageMat, self.MosaicSliceNumber)
                if image is None:
                    # 4维矩阵
                    image = np.repeat(_3d_image[:,:,:,np.newaxis], len(self), 3)
                else:
                    if _3d_image.shape != image.shape[:-1]:
                        raise Exception('维度不一致')
                    image[:, :, :, idx] = _3d_image

        elif self.dicom_image_type == DICOM_IMAGE_TYPE.DICOM_IMAGE_TYPE_2D:
            image = np.zeros((self.Rows, self.Columns, self.SliceNumber, self.Repetition))
            if self.is_second_arrangement_mode:
                for r in range(self.Repetition):
                    for n in range(self.SliceNumber):
                        instance_number = n * self.Repetition + r
                        image[:, :, n, r] = self[instance_number].ImageMat
            else:
                for r in range(self.Repetition):
                    for n in range(self.SliceNumber):
                        instance_number = r * self.SliceNumber + n
                        image[:, :, n, r] = self[instance_number].ImageMat

        elif self.dicom_image_type == DICOM_IMAGE_TYPE.DICOM_IMAGE_TYPE_3D:
            image = np.zeros((self.Rows, self.Columns, self.SliceNumber, self.Repetition))
            for no, _3d_image in enumerate(self):
                if self.is_third_arrangement_mode:
                    # 将第三种模式转换成第二种模式
                    image_mat = np.zeros(_3d_image.ImageMat.shape, dtype=_3d_image.ImageMat.dtype)
                    single_group = self.Repetition // 2
                    for i in range(0, _3d_image.ImageMat.shape[-1], 2):
                        image_mat[..., i] = _3d_image.ImageMat[...,i//2]
                        image_mat[..., i + 1] = _3d_image.ImageMat[...,single_group+i//2]
                    _3d_image.ImageMat = image_mat
                    self.is_third_arrangement_mode = False
                    self.is_second_arrangement_mode = True

                repetition = self.Repetition // len(self)
                if self.is_second_arrangement_mode:
                    for r in range(repetition):
                        for n in range(self.SliceNumber):
                            instance_number = n * self.Repetition + r
                            image[:, :, n, no * repetition + r] = _3d_image.ImageMat[...,instance_number]
                else:
                    for r in range(repetition):
                        for n in range(self.SliceNumber):
                            instance_number = r * self.SliceNumber + n
                            image[:, :, n, no * repetition + r] = _3d_image.ImageMat[...,instance_number]
        # 数值转换
        return image if image is None else (image * self.RescaleSlope + self.RescaleIntercept) * rescale

    def merge(self, other):
        if not isinstance(other, SeriesModel):
            raise Exception("only merge SeriesModel")
        if other.StudyInstanceUID != self.StudyInstanceUID:
            raise Exception('无法合并并不属于同一Study的序列')
        self.extend(other)
        self.Repetition += other.Repetition

    def __add__(self, other):
        return self.merge(other)

    def _calc_2d_affine(self):
        # 原点与方向
        orientation = self.ImageOrientationPatient
        position = self.ImagePositionPatient

        # 分辨率
        z_space = self.SpacingBetweenSlices
        x_space, y_space = self.PixelSpacing

        m = np.zeros((3, 3), dtype=float)
        m[0] = orientation[:3]
        m[1] = orientation[3:]
        val = sum(np.power(m[0], 2))
        if val > 0:
            val = 1 / np.sqrt(val)
            m[0] *= val
        else:
            m[0] = [1, 0, 0]

        val = sum(np.power(m[1], 2))
        if val > 0:
            val = 1 / np.sqrt(val)
            m[1] *= val
        else:
            m[1] = [0, 1, 0]

        # /* row 3 is the cross product of rows 1 and 2*/
        m[2, 0] = m[0, 1] * m[1, 2] - m[0, 2] * m[1, 1]
        m[2, 1] = m[0, 2] * m[1, 0] - m[0, 0] * m[1, 2]
        m[2, 2] = m[0, 0] * m[1, 1] - m[0, 1] * m[1, 0]

        # 转置阵
        m = m.T

        if np.linalg.det(m) < 0:
            m[:2] *= -1

        # 乘以分辨率
        diag_vox = np.array([[x_space, 0, 0], [0, y_space, 0], [0, 0, z_space]])
        m = np.matmul(m, diag_vox)

        m = np.c_[m, position]
        m = np.r_[m, [[0, 0, 0, 1]]]

        return m

    def calc_affine(self):
        m = self._calc_2d_affine()

        # mosaic的坐标系，需要进一步转换成2D的坐标系
        if self.dicom_image_type == DICOM_IMAGE_TYPE.DICOM_IMAGE_TYPE_MOSAIC:
            factor_x = factor_y = (self.Rows - (self.Columns / int(np.ceil(np.sqrt(self.SliceNumber))))) / 2.0
            m[:,3] += m[:,0] * factor_x + m[:,1] * factor_y


        # 要反转的是 slice direction todo：不一定是[2,2]???
        if self.z_reversed:
            Q = np.eye(4)
            Q[2, 2] *= -1
            m = np.matmul(m, Q)

        return m

def dicom_parse_series(dir: str, parse_after_hook=None) -> Dict[str, Dict[str, SeriesModel]]:
    study_list = {}
    for root, _, files in os.walk(dir):
        for fn in files:
            abs_fn = os.path.join(root, fn)
            try:
                ds = pydicom.read_file(abs_fn)
                ds_tags = InstanceModel(ds, abs_fn)

                if parse_after_hook != None:
                    # 执行hook
                    parse_after_hook(abs_fn, ds_tags)
            except InvalidDicomError:
                logger.debug(f'忽略文件：{abs_fn}')
                continue
            except PhotometricRGBError:
                logger.debug(f'忽略彩色图像：{abs_fn}')
                continue
            except Exception as e:
                logger.debug(f'{e}：{abs_fn}')
                continue
            if ds_tags.StudyInstanceUID not in study_list:
                study_list[ds_tags.StudyInstanceUID] = {}
            if ds_tags.SeriesInstanceUID not in study_list[ds_tags.StudyInstanceUID]:
                study_list[ds_tags.StudyInstanceUID][ds_tags.SeriesInstanceUID] = SeriesModel(ds_tags, ds)
            study_list[ds_tags.StudyInstanceUID][ds_tags.SeriesInstanceUID].append(ds_tags)

    for study in study_list.values():
        for series in study.values():
            series.finish_handle()
    return study_list



"""
修改了-modify（ReScaleIntercept)为永久是-5，数据本身不稳定
读取的数据的负数都会变成0

21年06月29日
姚泽山
"""

import pydicom
import os
import numpy as np
from .AnImageASLUtility import AnImageASLUtility
from . import AnImageDCMUtility as DCMUtil


class AnImageDataSorterGE7:  # FIXME：需要整理代码，整体按照行号227-260的格式修改 [姚 06-14]
    def __init__(self):
        self.ref_img = ''
        self.intr_fldr = ''
        self.n = 0
        self.mask = []
        self.cbf_ncorr = []
        self.raw_data_pwpd = []
        self.st = 0
        self.sl = 0
        self.ipp = []
        self.iop = []
        self.dmrow = 0
        self.folder_suffix = ''
        self.series_CORR = []
        self.series_STAN = []
        self.label_dur = []
        self.delay_time = []
        self.aslutil = AnImageASLUtility()

    def set_up_aslutil(self):
        image_data = pydicom.dcmread(self.ref_img)
        self.aslutil.set_vendor('ge')
        self.aslutil.set_series_type('7delay')
        self.aslutil.set_duration_list(self.label_dur)
        self.aslutil.set_delay_list(self.delay_time)
        self.aslutil.set_measure_list([1] * 7)
        self.aslutil.set_scan_parameter(image_data)

        if [0x0021, 0x104f] in image_data:
            self.n = image_data[0x0021, 0x104f].value
        else:
            raise Exception('slice number not find, will fix later')

        self.dmrow = image_data[0x0028, 0x0010].value
        self.st = image_data[0x0018, 0x0050].value
        self.sl = image_data[0x0020, 0x1041].value
        self.ipp = image_data[0x0020, 0x0032].value
        self.iop = image_data[0x0020, 0x0037].value

    def parse_m0_from_raw_folder(self, seriesm0):
        self.raw_data_pwpd = np.zeros((self.dmrow, self.dmrow, self.n, 9))
        for series_idx in range(9):
            for slice_idx in range(self.n):
                image_data = pydicom.dcmread(seriesm0[0][self.n * series_idx + slice_idx + 1])
                self.raw_data_pwpd[:, :, slice_idx, series_idx] = image_data.pixel_array

        self.raw_data_pwpd[self.raw_data_pwpd < 0] = 0
        self.aslutil.set_raw_data(self.raw_data_pwpd)
        self.aslutil.gen_m0()
        self.aslutil.gen_mask()

        output_array = self.aslutil.get_m0()[:, :, :, 0]
        os.makedirs(os.path.join(self.intr_fldr, 'asl-' + self.folder_suffix, 'm0'), exist_ok=True)
        series_param = [1, 1, 0]
        DCMUtil.create_dcm_series(output_array, self.ref_img, self.intr_fldr,
                                  'asl', 'm0', series_param, self.folder_suffix)
        DCMUtil.dcm_to_nii(self.intr_fldr, 'asl-' + self.folder_suffix, 'm0')

    def ge_calc_yao(self, seriesatt):
        att_array = 0 * self.aslutil.get_m0()[:, :, :, 0]
        for slice_idx in range(len(seriesatt[0]) - 1):
            image_data = pydicom.dcmread(seriesatt[0][slice_idx + 1])
            att_array[:, :, slice_idx] = image_data.pixel_array.astype(np.float32) / 1000

        self.aslutil.set_att(att_array)
        self.aslutil.gen_plds()
        self.aslutil.gen_cbfs_ge7delay(transit_correction=False)
        self.cbf_ncorr = self.aslutil.get_mcbf()
        self.aslutil.gen_cbfs_ge7delay(transit_correction=True)
        self.aslutil.gen_acbv()

    def generate_scbf(self):
        if self.series_CORR:
            if len(self.series_CORR[0]) > 100:
                output_array = np.zeros((self.dmrow, self.dmrow, self.n, 7))
                for series_idx in range(7):
                    for slice_idx in range(self.n):
                        image_data = pydicom.dcmread(self.series_CORR[0][self.n * series_idx + slice_idx + 1])
                        output_array[:, :, slice_idx, series_idx] = image_data.pixel_array
            else:
                output_array = self.aslutil.get_cbfs()
        else:
            output_array = self.aslutil.get_cbfs()
        output_array[output_array < 0] = 0

        for file_idx in range(7):
            os.makedirs(os.path.join(self.intr_fldr, 'asl-' + self.folder_suffix,
                                     'cbf{:d}'.format(file_idx+1)), exist_ok=True)
            series_param = [10 + file_idx + 1, 10, 0]
            DCMUtil.create_dcm_series(output_array[:, :, :, file_idx], self.ref_img, self.intr_fldr, 'asl',
                                      'cbf{:d}'.format(file_idx+1), series_param, self.folder_suffix)
            DCMUtil.dcm_to_nii(self.intr_fldr, 'asl-' + self.folder_suffix,
                               'cbf{:d}'.format(file_idx+1))

    def generate_mcbf_noncorr(self):
        target_files = ["mcbf", "att", "acbv", "nc_cbf"]
        scaling = [10, 1000, 1000, 10]

        for file_idx in range(len(target_files)):
            os.makedirs(os.path.join(self.intr_fldr, 'asl-' + self.folder_suffix,
                                     '{}'.format(target_files[file_idx])), exist_ok=True)
            if file_idx == 0:
                output_array = self._check_series_CORR()
            elif file_idx == 1:
                output_array = self.aslutil.get_att()
            elif file_idx == 2:
                output_array = self.aslutil.get_acbv()
            elif file_idx == 3:
                output_array = self._check_series_STAN()

            output_array[output_array < 0] = 0
            series_param = [20 + file_idx + 1, scaling[file_idx], 0]
            DCMUtil.create_dcm_series(output_array, self.ref_img, self.intr_fldr, 'asl',
                                      '{}'.format(target_files[file_idx]), series_param, self.folder_suffix)
            DCMUtil.dcm_to_nii(self.intr_fldr, 'asl-' + self.folder_suffix,
                               '{}'.format(target_files[file_idx]))

    def _check_series_CORR(self):
        if self.series_CORR:
            if 0 < len(self.series_CORR[0]) < 100:
                output_array = np.zeros((self.dmrow, self.dmrow, self.n))
                for slice_idx in range(self.n):
                    image_data = pydicom.dcmread(self.series_CORR[0][slice_idx + 1])
                    output_array[:, :, slice_idx] = image_data.pixel_array
            else:
                output_array = self.aslutil.get_mcbf()
        else:
            output_array = self.aslutil.get_mcbf()
        # output_array = self.aslutil.get_mcbf()
        return output_array

    def _check_series_STAN(self):
        if self.series_STAN:
            if 0 < len(self.series_STAN[0]) < 100:
                output_array = np.zeros((self.dmrow, self.dmrow, self.n))
                for slice_idx in range(self.n):
                    image_data = pydicom.dcmread(self.series_STAN[0][slice_idx + 1])
                    output_array[:, :, slice_idx] = image_data.pixel_array
            else:
                output_array = self.cbf_ncorr
        else:
            output_array = self.cbf_ncorr
        return output_array

        # imgs = aslutil.get_mcbf()
        # imgs[imgs < 0] = 0  # FIXME: 额外加的，可能对数值好一些，需要查看
        # self.mcbf = imgs
        # modify = 0
        # mcbf_factor = 1
        # if np.min(imgs) < 0:
        #     if (int(abs(np.min(imgs))) // 10 + 1) * 10 > modify:
        #         modify = (int(abs(np.min(imgs))) // 10 + 1) * 10
        # imgs[self.mask[:, :, :, 0] == 0] = -modify * mcbf_factor
        # for slice_idx in range(imgs.shape[2]):
        #     img_mcbf = (imgs[:, :, slice_idx] + modify) * mcbf_factor
        #     if [0x0028, 0x1052] in test:
        #         test[0x0028, 0x1052].value = -modify * mcbf_factor
        #     else:
        #         test.add_new([0x0028, 0x1052], 'DS', -modify * mcbf_factor)
        #     # if [0x0028, 0x1053] in test:
        #     #     test[0x0028, 0x1053].value = 0.1
        #     # else:
        #     #     test.add_new([0x0028, 0x1053], 'DS', 0.1)
        #     test[0x0020, 0x0013].value = slice_idx + 1
        #     test[0x0020, 0x1041].value = self.sl + slice_idx * self.st
        #     self.ipp[2] = '{:.4f}'.format(test[0x0020, 0x1041].value)
        #     test[0x0020, 0x0032].value = self.ipp
        #     test[0x0020, 0x0037].value = self.iop
        #     test = SorterUtil.modify_series_uid(test, 'asl-mcbf')
        #     test = SorterUtil.add_AnImage_tag(test, 'asl-mcbf')
        #     test = SorterUtil.get_and_set_bytetype(test, img_mcbf)
        #     test.save_as(os.path.join(self.intr_fldr, 'asl-' + i,
        #                               'mcbf',
        #                               '{:04d}.dcm'.format(slice_idx + 1)))
        #
        # make_nifti(os.path.join(self.intr_fldr, 'asl-' + i), 'mcbf')

    def Sorter(self, intermediate_folder, folder_suffix, seriesdic, delay_time, label_dur):
        # print(json.dumps(seriesdic, sort_keys=True, indent=4))
        if "ATT" in seriesdic and "RAW" in seriesdic:
            os.makedirs(os.path.join(self.intr_fldr, 'asl-' + folder_suffix), exist_ok=True)
            self.intr_fldr = intermediate_folder
            self.folder_suffix = folder_suffix
            self.ref_img = seriesdic["RAW"][0][1]

            self.label_dur = label_dur
            self.delay_time = delay_time

            self.set_up_aslutil()
            self.parse_m0_from_raw_folder(seriesdic['RAW'])
            self.ge_calc_yao(seriesdic["ATT"])
            if "CORR_CBF" in seriesdic.keys():
                self.series_CORR = seriesdic["CORR_CBF"]
            if "STAN_CBF" in seriesdic.keys():
                self.series_STAN= seriesdic["STAN_CBF"]

            self.generate_scbf()
            self.generate_mcbf_noncorr()

        else:
            raise Exception('ATT and/or RAW instance not found!')



from ._dicom import *
from .errors import DataQuantityTooManyError
from .ge_3d_pcasl import DataSorterGEPCASL
from .ge_easl import DataSorterGEeASL
from .noASL_MRI import *
from .philips_3d_pcasl import DataSorterPhilipsPCASL
from .siemens_3d_pasl import DataSorterSiemensPASL
from .siemens_3d_pcasl import DataSorterSiemensPCASL
from .siemens_3d_pcasl_old import DataSorterSiemensPCASL_old
from .uih_3d_pasl import DataSorterUIHPASL


def AnImageDataSorter(output_root: str, data_type, series_list, delay_time, delay_rep, label_dur, extra_factor=None,
                      default_value_lambda=None, calc_c_cbf=False):
    if default_value_lambda is None:
        default_value_lambda = lambda k, m: []
    try:
        if len(delay_time) == 0:
            delay_time = default_value_lambda(data_type, 0)
        if len(label_dur) == 0:
            label_dur = default_value_lambda(data_type, 1)
        if len(delay_rep) == 0:
            delay_rep = default_value_lambda(data_type, 2)
    except Exception as e:
        logger.warning(f'{data_type} 获取ASL参数失败 => {str(e)}')
    if data_type == MRI_Type.MRI_TYPE_3D_PASL_SIEMENS:
        return DataSorterSiemensPASL(data_type, output_root, delay_time, delay_rep, label_dur, series_list, extra_factor, calc_c_cbf)
    elif data_type in [MRI_Type.MRI_TYPE_5Delay_3D_PCASL_SIEMENS, MRI_Type.MRI_TYPE_1Delay_3D_PCASL_SIEMENS]:
        return DataSorterSiemensPCASL(data_type, output_root, delay_time, delay_rep, label_dur, series_list, calc_c_cbf)
    elif data_type == MRI_Type.MRI_TYPE_4Delay_3D_PCASL_SIEMENS_Old:
        return DataSorterSiemensPCASL_old(data_type, output_root, delay_time, delay_rep, label_dur, series_list, calc_c_cbf)
    elif data_type == MRI_Type.MRI_TYPE_3D_PCASL_PHILIPS:
        return DataSorterPhilipsPCASL(data_type, output_root, delay_time, delay_rep, label_dur, series_list, calc_c_cbf)
    elif data_type == MRI_Type.MRI_TYPE_3D_PCASL_GE:
        return DataSorterGEPCASL(data_type, output_root, delay_time, delay_rep, label_dur, series_list, calc_c_cbf)
    elif data_type == MRI_Type.MRI_TYPE_eASL_PCASL_GE:
        return DataSorterGEeASL(data_type, output_root, delay_time, delay_rep, label_dur, series_list, calc_c_cbf)
    elif data_type == MRI_Type.MRI_TYPE_3D_PASL_UIH:
        return DataSorterUIHPASL(data_type, output_root, delay_time, delay_rep, label_dur, series_list, 4, calc_c_cbf)
    elif data_type == MRI_Type.MRI_TYPE_T1:
        return DataSorterT1(data_type, output_root, series_list)
    elif data_type == MRI_Type.MRI_TYPE_T2:
        return DataSorterT2(data_type, output_root, series_list)
    elif data_type == MRI_Type.MRI_TYPE_T2_FLAIR:
        return DataSorterT2FLAIR(data_type, output_root, series_list)
    elif data_type == MRI_Type.MRI_TYPE_DWI:
        return DataSorterDWI(data_type, output_root, series_list)
    else:
        raise UnSupportDataTypeError(data_type)


def check_study_series(series_group_map: Dict[MRI_Type, List[SeriesModel]]):
    count = len(series_group_map)
    if MRI_Type.MRI_TYPE_UnKnown in series_group_map:
        count -= 1
    if MRI_Type.MRI_TYPE_T1 in series_group_map:
        count -= 1
    if MRI_Type.MRI_TYPE_T2 in series_group_map:
        count -= 1
    if MRI_Type.MRI_TYPE_T2_FLAIR in series_group_map:
        count -= 1
    if MRI_Type.MRI_TYPE_DWI in series_group_map:
        count -= 1
    if count > 1:
        raise DataQuantityTooManyError(count)
    # if count == 0:
    #     raise MissingSequenceError('ASL')

def ImageDataSort(src, out, delay_time, delay_rep, label_dur, extra_factor=None, default_value_lambda=None, calc_c_cbf=False):
    sorter_obj = []
    all_study_list = dicom_parse_series(src)

    # 只允许最多一个数据处理
    if len(all_study_list) > 1:
        raise DataQuantityTooManyError(len(all_study_list))

    # 划分可处理单元，每个study包含N个series_group
    for all_series_list in all_study_list.values():
        series_group_map = {}
        for series in all_series_list.values():
            if series.data_type not in series_group_map:
                series_group_map[series.data_type] = []
            series_group_map[series.data_type].append(series)

        # 检查数据中是否符合要求：只包含一组ASL数据
        check_study_series(series_group_map)

        for data_type, series_list in series_group_map.items():
            if data_type == MRI_Type.MRI_TYPE_UnKnown:
                continue
            try:
                sorter = AnImageDataSorter(out, data_type, series_list, delay_time, delay_rep, label_dur, extra_factor,
                                           default_value_lambda=default_value_lambda, calc_c_cbf=calc_c_cbf)
                logger.debug(f'{sorter.data_type} 开始清洗数据')
                sorter.Sorter()
                sorter_obj.append(sorter)
                logger.debug(f'{sorter.data_type} 清洗数据完成')
            except Exception as e:
                logger.error(f'{data_type} 清洗错误=> {e}')
                raise e

    return sorter_obj

if __name__ == '__main__':
    pass




"""
修改，允许输入飞利浦1、2、3延迟数据
修改了多余重复的代码


21年07月19日
姚泽山
"""

import pydicom
import os
from . import AnImageDataSorterUtility as SorterUtil
from . import AnImageDCMUtility as DCMUtil
import numpy as np


# 飞利浦
class AnImageDataSorterPhilips13:
    def __init__(self):
        self.intermediate_folder = ''

    def parse_m0_from_raw_folder(self, series):
        os.makedirs(os.path.join(self.intermediate_folder, 'asl', 'm0'), exist_ok=True)
        # image_data = pydicom.dcmread(self.seriesm0[2][1])
        # temporal_num = image_data[0x0020, 0x0105].value
        # dmrow = image_data[0x0028, 0x0010].value
        # v = np.zeros(((len(self.seriesm0) // 2), self.dmrow, self.dmrow))

        # 查看飞利浦M0的数据格式

        if "M0" in series.keys():
            m0keys = sorted(list(series["M0"][0].keys()))
            img = pydicom.dcmread(series["M0"][0][m0keys[1]], stop_before_pixels=True)
            first_slice_loc = img[0x0020, 0x1041].value
            img = pydicom.dcmread(series["M0"][0][m0keys[2]], stop_before_pixels=True)
            second_slice_locA = img[0x0020, 0x1041].value
            img = pydicom.dcmread(series["M0"][0][m0keys[len(series["M0"][0]) // 2 + 1]], stop_before_pixels=True)
            second_slice_locB = img[0x0020, 0x1041].value

            for slice_idx in range(len(series["M0"][0]) // 2):
                if first_slice_loc == second_slice_locA:
                    image_data = pydicom.dcmread(series["M0"][0][m0keys[slice_idx * 2 + 1]])
                elif first_slice_loc == second_slice_locB:
                    image_data = pydicom.dcmread(series["M0"][0][m0keys[slice_idx + 1]])
                image_data = SorterUtil.modify_series_uid(image_data, 'asl-m0', None)
                image_data = SorterUtil.add_AnImage_tag(image_data, 'asl-m0')
                image_data[0x0020, 0x0013].value = slice_idx + 1
                image_data.save_as(
                    os.path.join(self.intermediate_folder, 'asl', 'm0', '{:04d}.dcm'.format(slice_idx + 1)))
        else:
            m0keys = sorted(list(series["CBF"][0].keys()))
            for slice_idx in range(len(series["CBF"][0]) - 1):
                image_data = pydicom.dcmread(series["CBF"][0][m0keys[slice_idx + 1]])
                image_data = SorterUtil.modify_series_uid(image_data, 'asl-m0', None)
                image_data = SorterUtil.add_AnImage_tag(image_data, 'asl-m0')
                image_data[0x0020, 0x0013].value = slice_idx + 1
                image_data.save_as(
                    os.path.join(self.intermediate_folder, 'asl', 'm0', '{:04d}.dcm'.format(slice_idx + 1)))

        DCMUtil.dcm_to_nii(self.intermediate_folder, 'asl', 'm0')

    def parse_cbf_from_archive_folder(self, series):
        seriescbf = series["CBF"]
        for series_idx in range(len(seriescbf)):
            os.makedirs(os.path.join(self.intermediate_folder, 'asl', 'cbf{:d}'.format(series_idx + 1)), exist_ok=True)

            if len(seriescbf) > 0 and series_idx != 0:
                os.makedirs(os.path.join(self.intermediate_folder, 'asl',
                                         'cbf{:d}-{:d}'.format(series_idx + 1, series_idx)), exist_ok=True)

            test = pydicom.dcmread(seriescbf[series_idx][2])
            series = np.zeros((len(seriescbf[series_idx]), test[0x0028, 0x0010].value, test[0x0028, 0x0011].value))
            for slice_idx in range(len(seriescbf[series_idx]) - 1):
                image_data = pydicom.dcmread(seriescbf[series_idx][slice_idx + 1])
                array_data = image_data.pixel_array
                image_data = SorterUtil.modify_series_uid(image_data, 'asl-cbf{:d}'.format(series_idx + 1), None)
                image_data = SorterUtil.add_AnImage_tag(image_data, 'asl-cbf{:d}'.format(series_idx + 1))
                image_data.save_as(os.path.join(self.intermediate_folder, 'asl',
                                                'cbf{:d}'.format(series_idx + 1), '{:04d}.dcm'.format(slice_idx + 1)))
                series[image_data[0x0020, 0x0013].value - 1, :, :] = array_data

                if len(seriescbf) > 0 and series_idx != 0:
                    diffdata = array_data - series2[image_data[0x0020, 0x0013].value - 1, :, :]
                    image_data.PixelData = diffdata.astype(np.uint16)
                    image_data.remove_private_tags()
                    image_data = SorterUtil.modify_series_uid(image_data,
                                                              'asl-cbf{:d}-{:d}'.format(series_idx + 1, series_idx),
                                                              None)
                    image_data = SorterUtil.add_AnImage_tag(image_data,
                                                            'asl-cbf{:d}-{:d}'.format(series_idx + 1, series_idx))
                    image_data.save_as(os.path.join(self.intermediate_folder, 'asl',
                                                    'cbf{:d}-{:d}'.format(series_idx + 1, series_idx),
                                                    '{:04d}.dcm'.format(image_data[0x0020, 0x0013].value)))
            series2 = series
            DCMUtil.dcm_to_nii(self.intermediate_folder, 'asl', 'cbf{:d}'.format(series_idx + 1))
            if len(seriescbf) > 0 and series_idx != 0:
                DCMUtil.dcm_to_nii(self.intermediate_folder, 'asl', 'cbf{:d}-{:d}'.format(series_idx + 1, series_idx))

    def Sorter(self, intermediate_folder, series):
        self.intermediate_folder = intermediate_folder
        if not os.path.exists(os.path.join(self.intermediate_folder, 'asl')):
            os.mkdir(os.path.join(self.intermediate_folder, 'asl'))
        self.parse_m0_from_raw_folder(series)
        self.parse_cbf_from_archive_folder(series)




from .old.AnImageDataSorter import ImageDataSort as ImageDataSort1
from .restructure import ImageDataSort as ImageDataSort2

def ImageDataSort(raw_folder, output_path, delay_time, delay_rep, label_dur, pasl_extra_factor, is_restructure=False,
                  calc_c_cbf=False):
    if is_restructure:
        return ImageDataSort2(raw_folder, output_path, delay_time, delay_rep, label_dur, pasl_extra_factor, calc_c_cbf=calc_c_cbf)
    else:
        return ImageDataSort1(raw_folder, output_path, delay_time, delay_rep, label_dur, pasl_extra_factor, zip=False,
                              debugmode=False, modify=0)



class DataSortException(Exception):
    errno = 20000
    def __init__(self, msg):
        super().__init__(msg)

class DicomMissingTagError(DataSortException):
    errno = 20001
    def __init__(self, key: tuple) -> None:
        super().__init__('Dicom Missing Tag (0x%04x, 0x%04x)'%(key[0], key[1]))

class CsaHeaderMissingKeyError(DataSortException):
    errno = 20002
    def __init__(self, key) -> None:
        super().__init__(f'Siemens CSA Header Missing Key {key}')

class UnSupportDataTypeError(DataSortException):
    errno = 20003
    def __init__(self, data_type) -> None:
        super().__init__(f'UnSupport DataType {data_type}')

class PhotometricRGBError(DataSortException):
    errno = 20004
    def __init__(self) -> None:
        super().__init__(f'RGB ignore')

class MissingSequenceError(DataSortException):
    errno = 20005
    def __init__(self, sequence_name: str) -> None:
        super().__init__(f'Missing Sequence {sequence_name}')

class MissingSliceError(DataSortException):
    errno = 20006
    def __init__(self, missing_number: int) -> None:
        super().__init__(f'Missing Slice No.{missing_number}')

class SliceLessError(DataSortException):
    errno = 20007
    # 切片数量太少
    def __init__(self, slice_number: int) -> None:
        super().__init__(f'The layer number of too little {slice_number}')

class SpaceIsDifferentError(DataSortException):
    errno = 20008
    # 空间不一致
    def __init__(self, *args) -> None:
        super().__init__(f'{"、".join([str(i) for i in args])} In the same space')

class DiffusionBValueMissingError(DataSortException):
    errno = 20009
    def __init__(self) -> None:
        super().__init__(f'DiffusionBValue Missing')

class UnidentifiedDicomImageTypeError(DataSortException):
    errno = 20010
    # DICOM 图像类型未明确
    def __init__(self) -> None:
        super().__init__(f'Dicom image type is unidentified')

class UnidentifiedOrientationError(DataSortException):
    errno = 20011
    # 方位未明确
    def __init__(self, affine) -> None:
        super().__init__(f'Orientation is unidentified {affine}')

class NotFoundSeriesError(DataSortException):
    errno = 20012
    def __init__(self) -> None:
        super().__init__(f'Not found Series')

class DelayNotEqualSeriesNumberError(DataSortException):
    errno = 20013
    def __init__(self, delay_count: int, series_count) -> None:
        super().__init__(f'PLD输入数量错误，请输入{series_count}个时间点')

class DelayNotEqualRepetitionError(DataSortException):
    errno = 20014
    def __init__(self, delay_count: int, repetition):
        super().__init__(f'The delay must be equal to the number of repetitions：Delay[{delay_count}]，Repetition[{repetition}]')

class ControlNotEqualLabelError(DataSortException):
    errno = 20015
    def __init__(self, control_count: int, label_count: int):
        super().__init__(f'Not Equal：control_count[{control_count}]，label_count[{label_count}]')

class SeriesError(DataSortException):
    errno = 20016
    def __init__(self, msg):
        super().__init__(msg)

class DataQuantityTooManyError(DataSortException):
    errno = 20017
    def __init__(self, c: int):
        super().__init__(f'The amount of data should be less than 1, current[{c}]')

class SeriesTooMany(DataSortException):
    errno = 20018
    def __init__(self, msg):
        super().__init__(msg)


import os
import time

import SimpleITK as sitk
import nibabel
import numpy as np
import pydicom
from nibabel.nicom import csareader
from pydicom._storage_sopclass_uids import SecondaryCaptureImageStorage
from pydicom.uid import generate_uid

'''
本文件为测试demo
'''

def create_dicom_forsitk(image:np.ndarray, position, orientation, slice_location, ref: pydicom.Dataset):
    image = sitk.GetImageFromArray(image.astype(np.uint16))
    image.SetMetaData('0020|0037', '\\'.join([str(i) for i in orientation]))
    image.SetMetaData('0020|0032', '\\'.join([str(i) for i in position]))
    image.SetMetaData('0020|1041', str(slice_location))


    image.SetMetaData('0020|000E', ref.SeriesInstanceUID)
    image.SetMetaData('0020|000d', ref.StudyInstanceUID)
    uid = generate_uid()
    image.SetMetaData('0002|0003', uid)
    image.SetMetaData('0008|0018', uid)
    image.SetMetaData('0002|0002', SecondaryCaptureImageStorage)
    image.SetMetaData('0008|0016', SecondaryCaptureImageStorage)
    image.SetMetaData('0008|0060', 'MR')
    image.SetMetaData('0008|0030', ref.StudyTime)
    image.SetMetaData('0008|0020', ref.StudyDate)
    image.SetMetaData('0028|0030', '\\'.join([str(i) for i in ref.PixelSpacing]))
    image.SetMetaData('0018|2010', '\\'.join([str(i) for i in ref.PixelSpacing]))
    image.SetMetaData('0018|0088', str(ref.SpacingBetweenSlices))
    # image.SetMetaData('0028|1052', str(-1))
    # image.SetMetaData('0028|1053', str(0.1))
    return image


def demosaic(mosaic_mat: np.ndarray, slice_num: int):
    n_rows = n_cols = int(np.ceil(np.sqrt(slice_num)))
    img_width, img_height = mosaic_mat.shape[1], mosaic_mat.shape[0]
    img_width //= n_cols
    img_height //= n_rows

    _3d_mat = np.zeros((img_height, img_width, slice_num))
    for i in range(slice_num):
        x_start = i % n_cols * img_width
        y_start = i // n_cols * img_height
        _3d_mat[::,:,i] = mosaic_mat[y_start:y_start+img_height, x_start:x_start+img_width]

    return _3d_mat


def nii_flipX(img: np.ndarray, affine: np.ndarray):
    s = affine[:3,:3]

    # 翻转后的偏移量
    v = [img.shape[0]-1, 0, 0, 1]
    v = np.matmul(affine, v)

    # 翻转后的方向
    mFlipZ = np.eye(3)
    mFlipZ[0,0] *= -1
    s = np.matmul(s, mFlipZ)

    # 切片方向与偏移量组合在一起
    s = np.c_[s, v[:3]]
    s = np.r_[s, [affine[3]]]
    return s

def nii_flipY(img: np.ndarray, affine: np.ndarray):
    s = affine[:3,:3]

    # 翻转后的偏移量
    v = [0, img.shape[1]-1, 0, 1]
    v = np.matmul(affine, v)

    # 翻转后的方向
    mFlipY = np.eye(3)
    mFlipY[1,1] *= -1
    s = np.matmul(s, mFlipY)

    # 切片方向与偏移量组合在一起
    s = np.c_[s, v[:3]]
    s = np.r_[s, [affine[3]]]
    return s

def nii_flipZ(img: np.ndarray, affine: np.ndarray):
    s = affine[:3,:3]

    # 翻转后的偏移量
    v = [0, 0, img.shape[2]-1, 1]
    v = np.matmul(affine, v)

    # 翻转后的方向
    mFlipZ = np.eye(3)
    mFlipZ[2,2] *= -1
    s = np.matmul(s, mFlipZ)

    # 切片方向与偏移量组合在一起
    s = np.c_[s, v[:3]]
    s = np.r_[s, [affine[3]]]
    return s


def siemens_affine(dataset):
    # 原点与方向
    orientation = dataset[0x0020, 0x0037].value
    position = dataset[0x0020, 0x0032].value

    # 分辨率
    z_space = dataset[0x0018, 0x0050].value
    x_space, y_space = dataset[0x0028, 0x0030].value

    m = np.zeros((3, 3), dtype=float)
    m[0] = orientation[:3]
    m[1] = orientation[3:]
    val = sum(np.power(m[0], 2))
    if val > 0:
        val = 1 / np.sqrt(val)
        m[0] *= val
    else:
        m[0] = [1, 0, 0]

    val = sum(np.power(m[1], 2))
    if val > 0:
        val = 1 / np.sqrt(val)
        m[1] *= val
    else:
        m[1] = [0, 1, 0]

    # /* row 3 is the cross product of rows 1 and 2*/
    m[2, 0] = m[0, 1] * m[1, 2] - m[0, 2] * m[1, 1]
    m[2, 1] = m[0, 2] * m[1, 0] - m[0, 0] * m[1, 2]
    m[2, 2] = m[0, 0] * m[1, 1] - m[0, 1] * m[1, 0]

    # 转置阵
    m = m.T

    if np.linalg.det(m) < 0:
        m[:2] *= -1

    # 乘以分辨率
    diag_vox = np.array([[x_space, 0, 0], [0, y_space, 0], [0, 0, z_space]])
    m = np.matmul(m, diag_vox)

    m = np.c_[m, position]
    m = np.r_[m, [[0, 0, 0, 1]]]

    return m


def siemens_mosaic_affine(dataset):
    m = siemens_affine(dataset)

    orientation = dataset[0x0020, 0x0037].value


    # mosaic图像尺寸
    row = dataset[0x0028, 0x0010].value
    col = dataset[0x0028, 0x0011].value

    _3d_slice_img = demosaic(dataset.pixel_array, slice_num)

    # 读取西门子CSA头
    csa_image = csareader.get_csa_header(dataset, 'image')
    sliceNormV = csa_image['tags']['SliceNormalVector']['items']

    factor_x = factor_y = (row - (row / int(np.ceil(np.sqrt(slice_num))))) / 2.0

    m[:,3] += m[:,0] * factor_x + m[:,1] * factor_y
    m[:2] *= -1

    slice_norm_mat = np.zeros((3, 3), dtype=float)
    slice_norm_mat[:, 0] = orientation[:3]
    slice_norm_mat[:, 1] = orientation[3:]
    slice_norm_mat[:, 2] = sliceNormV

    # 法向量相反，Z序需要翻转
    if np.linalg.det(slice_norm_mat) < 0:
        print('reversed')
        Q = np.eye(4)
        Q[2, 2] *= -1
        m = np.matmul(m, Q)



    return _3d_slice_img, m

def siemens_2d_affine(dataset):
    m = siemens_affine(dataset)
    m[:2] *= -1

    return None, m

def g_slice_location(first_slice_location, z_space, slice_num):
    cur_slice_location = first_slice_location
    for i in range(slice_num):
        yield cur_slice_location
        cur_slice_location += z_space

def g_position(m, slice_num):
    for i in range(slice_num):
        yield m[:3, 3] + i * m[:3, 2]

if __name__ == '__main__':
    dcm_fn = r'D:\WorkDir\cloud\dicom_normalization\dicom_demo\Mosaic\reverse\452.dcm'
    # dcm_fn = r'.\dicom_demo\Mosaic\reverse\M0\452.dcm'

    # dcm_fn = os.path.join(dcm_fn, '3179cb6b-b7e1e17c-6daa556e-fba45da6-86b5dbc0.dcm')

    dataset = pydicom.read_file(dcm_fn)
    slice_num = dataset[0x0019, 0x100a].value
    if slice_num > 0:
        _3d_img, affine = siemens_mosaic_affine(dataset)
        affine[:2] *= -1
        first_slice_location = dataset[0x0020, 0x1041].value
        normal_direction = affine[2,2] / abs(affine[2,2])
        ori = dataset[0x0020, 0x0037].value
        z_space = dataset[0x0018, 0x0050].value

        # plt.imshow(_3D_to_2D(_3d_img, 6, 6))
        # plt.show()

        writer = sitk.ImageFileWriter()
        writer.KeepOriginalImageUIDOn()

        i = 0
        start = time.time()
        for position, slice_location in zip(g_position(affine, _3d_img.shape[2]), g_slice_location(first_slice_location, normal_direction * z_space, _3d_img.shape[2])):



            image = create_dicom_forsitk(_3d_img[:, :, i], position, ori, slice_location, dataset)
            image.SetMetaData('0020|0013', str(i + 1))
            writer.SetFileName(os.path.join(r'D:\WorkDir\cloud\dicom_normalization\dicom_demo\Mosaic\reverse\split', 'MR%06d.dcm' % (i)))
            writer.Execute(image)

            i+=1
        print(time.time() - start)

    else:
        _3d_img, affine = siemens_2d_affine(dataset)

    _3d_img = _3d_img.swapaxes(0, 1)
    nibabel.Nifti1Image(_3d_img, affine).to_filename(r'D:\WorkDir\cloud\dicom_normalization\dicom_demo\Mosaic\reverse\a.nii.gz')

    #
    # print(nibabel.aff2axcodes(affine))
    # print(affine)






from ._dicom import *
from ._dicom_util import write_dicom_series
from .data_sorter_base import DataSorterASLBase
from .errors import *


class DataSorterGEPCASL(DataSorterASLBase):
    def __init__(self, data_type, output_root, delay_time, delay_rep, label_dur, series_list, calc_c_cbf):
        DataSorterASLBase.__init__(self, data_type, output_root, delay_time, delay_rep, label_dur, series_list, calc_c_cbf)
        self.PWPD = [] # 灌注图
        self.CBF = [] # CBF参数图

        # 区别PWPD与CBF图
        for series in self.series_list:
            series_description = series.SeriesDescription.lower()
            if not isinstance(series, SeriesModel):
                continue
            elif series_description.find('3d') > -1 and series_description.find('asl') > -1:
                self.PWPD.append(series)
            elif series_description.find('cerebral blood flow') > -1 or series_description.find('cbf') > -1:
                self.CBF.append(series)

    def Sorter(self):
        super(DataSorterGEPCASL, self).Sorter()

        # 验证序列
        self.validate_series()

        # 多延迟存在于不同序列需要重新排序
        self.PWPD.sort(key=lambda x: x.SeriesNumber)
        self.CBF.sort(key=lambda x: x.SeriesNumber)

        self.generate_cbf_3d_pcasl()

    def validate_series(self):
        delay_time = []
        for i in self.PWPD:
            delay_time.append(i.get_tag((0x0018, 0x0082)))
        if len(delay_time) > 0:
            self.delay_time = [float(i) / 1000 for i in delay_time]
        logger.info(f'3D ASL 使用PLD为：{self.delay_time}')

        delay_count = len(self.delay_time)
        cbf_count = len(self.CBF)
        pwpd_count = len(self.PWPD)
        if cbf_count == 0 and pwpd_count == 0:
            raise NotFoundSeriesError()
        elif cbf_count == 0:
            logger.info('未发现CBF图，将由3D ASL计算生成')
        elif pwpd_count == 0:
            logger.warning('未发现PWPD，可能影响配准的效果')
        else:
            # 两种序列都存在，PWPD数量应为CBF两倍
            for i, j in zip(self.PWPD, self.CBF):
                if i.Repetition != 2 * j.Repetition:
                    raise SliceLessError(i.SliceNumber)

        if cbf_count != delay_count and cbf_count != 0:
            raise DelayNotEqualSeriesNumberError(delay_count, cbf_count)
        if pwpd_count != delay_count and pwpd_count != 0:
            raise DelayNotEqualSeriesNumberError(delay_count, pwpd_count)

    def set_scan_parameter(self, series_model: SeriesModel):
        self.number_of_averages = series_model.get_tag((0x0018, 0x0083), 1)
        self.magnetic_field_strength = series_model.get_tag((0x0018, 0x0087), 1)
        self.labeling_duration = series_model.get_tag((0x0043, 0x10A5), 1500) / 1000

    def generate_cbf_3d_pcasl(self):
        m0_image = None
        pw_image = None
        ref_series = None

        cbf_image = None
        if len(self.CBF) != 0:
            cbf_image = np.concatenate([i.load() for i in self.CBF], axis=3)
            ref_series = self.CBF[0]

        if len(self.PWPD) != 0:
            pwpd_image = np.concatenate([i.load() for i in self.PWPD], axis=3)
            pwpd_image[pwpd_image < 0] = 0
            m0_image = pwpd_image[:, :, :, 1::2]
            pw_image = pwpd_image[:, :, :, 0::2]
            ref_series = self.PWPD[0]

        if cbf_image is None:
            cbf_image = np.zeros(pw_image.shape)
            self.set_scan_parameter(self.PWPD[0])
            mask = self._gen_mask(m0_image)
            # 计算 CBF
            for i in range(pw_image.shape[3]):
                c_lambda = 0.9
                c_st = 2
                c_t1t = 1.2
                pld = self.delay_time[i]
                c_t1b = 1.4  # 1.5T磁场的参数
                if self.magnetic_field_strength == 3:
                    c_t1b = 1.6  # 3.0T磁场的参数
                c_epsilon = 0.6
                c_tau = self.labeling_duration
                nex = self.number_of_averages
                sf = 32  # 在DicomTag里找不到这个信息
                physiology_related = (1 - np.exp(-c_st / c_t1t)) * np.exp(pld / c_t1b) / c_t1b / (
                        1 - np.exp(-c_tau / c_t1b))
                scan_parameter_related = c_lambda / c_epsilon / sf / nex
                constant = 6000 / 2 * physiology_related * scan_parameter_related
                m0 = m0_image[:, :, :, i]
                m0[m0 < 1] = 1
                cbf = np.divide(pw_image[:, :, :, i], m0) * constant
                cbf[mask[..., i] == 0] = 0
                cbf_image[..., i] = cbf

        elif m0_image is None:
            # 无M0使用CBF替代
            m0_image = cbf_image

        cur_series_number = SeriesNumberStartWith.ASL
        write_dicom_series(m0_image[...,0], ref_series, 'asl-m0', cur_series_number, os.path.join(self.output_root, 'asl', 'm0'))
        for i, d in enumerate(self.delay_time):
            cur_series_number += 1
            write_dicom_series(cbf_image[:,:,:,i] * 10, ref_series, f'asl-cbf{i+1}', cur_series_number,
                               os.path.join(self.output_root, 'asl', f'cbf{i+1}'), slope=0.1)

        super().calc_correct_CBF(cbf_image, ref_series, cur_series_number + 1)

    def _gen_mask(self, m0: np.ndarray):
        mask = np.zeros(m0.shape)
        for i in range(m0.shape[3]):
            pd_data = m0[:, :, :, i]
            threshold = 0.5 * np.sum(np.multiply(pd_data, pd_data)) / np.sum(pd_data)
            mask[:, :, :, i][pd_data >= threshold] = 1
        return mask



import os
import pydicom
import numpy as np
import math
from . import AnImageDataSorterUtility as SorterUtil
import dicom2nifti
from .AnImageASLUtility import AnImageASLUtility


def create_path(target_path):
    if not os.path.exists(target_path):
        os.mkdir(target_path)


def make_nifti(pre_path, input_name):
    dicom2nifti.dicom_series_to_nifti(os.path.join(pre_path, input_name),
                                      os.path.join(pre_path, input_name, input_name + '.nii.gz'), reorient_nifti=True)


# 西门子1，5延迟处理
class AnImageDataSorterSiemens_UCLA:

    def __init__(self):
        self.intermediate_folder = ''
        self.seriesdelay = 0
        self.debugMode = False
        self.formatpack = []
        self.row = 0
        self.col = 0

    def parse_m0_from_raw_folder(self, seriesm0):
        create_path(os.path.join(self.intermediate_folder, 'asl', 'm0'))
        try:
            image_data = pydicom.dcmread(seriesm0[1])
        except:
            image_data = pydicom.dcmread([i for i in seriesm0.values()][1])
        # m0 = image_data.pixel_array
        self.row = image_data[0x0028, 0x0010].value
        self.col = image_data[0x0028, 0x0011].value
        # mask = np.zeros((self.row, self.col))
        # threshold = 0.15
        # mask[m0 < m0.max() * threshold] = 1
        slice_num = image_data[0x0019, 0x100A].value
        sn = math.ceil(np.sqrt(slice_num))
        dmrow = image_data[0x0028, 0x0010].value // sn
        siemensst = image_data[0x0018, 0x0050].value
        siemenssl = image_data[0x0020, 0x1041].value
        siemensipp = image_data[0x0020, 0x0032].value
        siemensiop = [1, 0, 0, 0, 1, 0]
        self.formatpack = [slice_num, sn, dmrow, siemensst, siemenssl, siemensipp, siemensiop]
        SorterUtil.demosaic(image_data, image_data.pixel_array, 'm0', self.intermediate_folder, self.formatpack)

    def parse_cbf_from_archive_folder(self, seriesmcbf, seriesatt, seriesacbv, seriescbf):
        # acbv_factor=1.65
        # att_factor=9
        # mcbf_factor=0.56
        acbv_factor = 1
        att_factor = 1
        mcbf_factor = 1
        # mcbf_factor=1
        if self.seriesdelay == 5:
            create_path(os.path.join(self.intermediate_folder, 'asl', 'mcbf'))
            create_path(os.path.join(self.intermediate_folder, 'asl', 'att'))
            create_path(os.path.join(self.intermediate_folder, 'asl', 'acbv'))

            test = pydicom.dcmread(seriesmcbf[1])
            studytime = test[0x0008, 0x0030].value
            for file_idx in range(len(seriesmcbf) - 1):
                image_data = pydicom.dcmread((seriesmcbf[file_idx + 1]))
                image_data[0x0008, 0x0030].value = studytime
                array = image_data.pixel_array * mcbf_factor
                image_data.PixelData = array.astype(np.uint16)
                image_data[0x0020, 0x0013].value = image_data[0x0020, 0x0013].value + 1
                image_data.add_new([0x0020, 0x1041], 'DS', self.formatpack[4] + file_idx * self.formatpack[3])
                self.formatpack[5][2] = self.formatpack[4] + file_idx * self.formatpack[3]
                image_data.add_new([0x0020, 0x0032], 'DS', self.formatpack[5])
                image_data.add_new([0x0020, 0x0037], 'DS', self.formatpack[6])
                if [0x0028, 0x1053] in image_data:
                    image_data[0x0028, 0x1053].value = 0.1
                else:
                    image_data.add_new([0x0028, 0x1053], 'DS', 0.1)
                image_data = SorterUtil.add_AnImage_tag(image_data, 'asl-mcbf')
                image_data = SorterUtil.modify_series_uid(image_data, 'asl-mcbf')
                image_data.save_as(os.path.join(self.intermediate_folder, 'asl', 'mcbf',
                                                '{:04d}.dcm'.format(int(image_data[0x0020, 0x0013].value))))

            make_nifti(os.path.join(self.intermediate_folder, 'asl'), 'mcbf')

            test = pydicom.dcmread(seriesatt[1])
            studytime = test[0x0008, 0x0030].value
            for file_idx in range(len(seriesatt) - 1):
                image_data = pydicom.dcmread((seriesatt[file_idx + 1]))
                image_data[0x0008, 0x0030].value = studytime
                array = image_data.pixel_array * att_factor
                image_data[0x0020, 0x0013].value = image_data[0x0020, 0x0013].value + 1
                image_data.add_new([0x0020, 0x1041], 'DS', self.formatpack[4] + file_idx * self.formatpack[3])
                self.formatpack[5][2] = self.formatpack[4] + file_idx * self.formatpack[3]
                image_data.add_new([0x0020, 0x0032], 'DS', self.formatpack[5])
                image_data.add_new([0x0020, 0x0037], 'DS', self.formatpack[6])
                if [0x0028, 0x1053] in image_data:
                    image_data[0x0028, 0x1053].value = 0.01
                else:
                    image_data.add_new([0x0028, 0x1053], 'DS', 0.01)
                image_data.PixelData = array.astype(np.uint16)
                image_data = SorterUtil.add_AnImage_tag(image_data, 'asl-att')
                image_data = SorterUtil.modify_series_uid(image_data, 'asl-att')
                image_data.save_as(os.path.join(self.intermediate_folder, 'asl', 'att',
                                                '{:04d}.dcm'.format(int(image_data[0x0020, 0x0013].value))))

            make_nifti(os.path.join(self.intermediate_folder, 'asl'), 'att')

            test = pydicom.dcmread(seriesacbv[1])
            studytime = test[0x0008, 0x0030].value
            for file_idx in range(len(seriesacbv) - 1):
                image_data = pydicom.dcmread((seriesacbv[file_idx + 1]))
                image_data[0x0008, 0x0030].value = studytime
                image_data[0x0020, 0x0013].value = image_data[0x0020, 0x0013].value + 1
                image_data.add_new([0x0020, 0x1041], 'DS', self.formatpack[4] + file_idx * self.formatpack[3])
                self.formatpack[5][2] = self.formatpack[4] + file_idx * self.formatpack[3]
                image_data.add_new([0x0020, 0x0032], 'DS', self.formatpack[5])
                image_data.add_new([0x0020, 0x0037], 'DS', self.formatpack[6])
                if [0x0028, 0x1053] in image_data:
                    image_data[0x0028, 0x1053].value = 0.001
                else:
                    image_data.add_new([0x0028, 0x1053], 'DS', 0.001)
                array = image_data.pixel_array * acbv_factor
                image_data.PixelData = array.astype(np.uint16)
                image_data = SorterUtil.add_AnImage_tag(image_data, 'asl-acbv')
                image_data = SorterUtil.modify_series_uid(image_data, 'asl-acbv')
                image_data.save_as(os.path.join(self.intermediate_folder, 'asl', 'acbv',
                                                '{:04d}.dcm'.format(int(image_data[0x0020, 0x0013].value))))

            make_nifti(os.path.join(self.intermediate_folder, 'asl'), 'acbv')

        for series_idx in range(self.seriesdelay):
            create_path(os.path.join(self.intermediate_folder, 'asl', 'cbf{:d}'.format(series_idx + 1)))
            if series_idx != 0:
                create_path(
                    os.path.join(self.intermediate_folder, 'asl', 'cbf{:d}-{:d}'.format(series_idx + 1, series_idx)))

            test = pydicom.dcmread(seriescbf[series_idx][1])
            studytime = test[0x0008, 0x0030].value
            series = np.zeros((len(seriescbf[series_idx]) - 1, test[0x0028, 0x0010].value, test[0x0028, 0x0011].value))
            for slice_idx in range(len(seriescbf[series_idx]) - 1):
                image_data = pydicom.dcmread(seriescbf[series_idx][slice_idx + 1])
                image_data[0x0008, 0x0030].value = studytime
                array = image_data.pixel_array
                array2 = image_data.pixel_array
                image_data[0x0020, 0x0013].value = image_data[0x0020, 0x0013].value + 1
                image_data.PixelData = array.astype(np.uint16)
                image_data.add_new([0x0020, 0x1041], 'DS', self.formatpack[4] + slice_idx * self.formatpack[3])
                self.formatpack[5][2] = self.formatpack[4] + slice_idx * self.formatpack[3]
                image_data.add_new([0x0020, 0x0032], 'DS', self.formatpack[5])
                image_data.add_new([0x0020, 0x0037], 'DS', self.formatpack[6])
                if [0x0028, 0x1053] in image_data:
                    image_data[0x0028, 0x1053].value = 0.1
                else:
                    image_data.add_new([0x0028, 0x1053], 'DS', 0.1)
                image_data = SorterUtil.add_AnImage_tag(image_data, 'asl-cbf{:d}'.format(series_idx + 1))
                image_data = SorterUtil.modify_series_uid(image_data, 'asl-cbf{:d}'.format(series_idx + 1))
                image_data.save_as(os.path.join(self.intermediate_folder, 'asl',
                                                'cbf{:d}'.format(series_idx + 1),
                                                '{:04d}.dcm'.format(int(image_data[0x0020, 0x0013].value))))
                series[image_data[0x0020, 0x0013].value - 1, :, :] = array2

                if series_idx != 0:
                    diffdata = array2 - series2[image_data[0x0020, 0x0013].value - 1, :, :]
                    image_data.remove_private_tags()
                    image_data.PixelData = diffdata.astype(np.uint16)
                    if [0x0028, 0x1053] in image_data:
                        image_data[0x0028, 0x1053].value = 0.1
                    else:
                        image_data.add_new([0x0028, 0x1053], 'DS', 0.1)
                    image_data = SorterUtil.add_AnImage_tag(image_data,
                                                            'asl-cbf{:d}-{:d}'.format(series_idx + 1, series_idx))
                    image_data = SorterUtil.modify_series_uid(image_data,
                                                              'asl-cbf{:d}-{:d}'.format(series_idx + 1, series_idx))
                    image_data.save_as(os.path.join(self.intermediate_folder, 'asl',
                                                    'cbf{:d}-{:d}'.format(series_idx + 1, series_idx),
                                                    '{:04d}.dcm'.format(int(image_data[0x0020, 0x0013].value))))

            series2 = series
            make_nifti(os.path.join(self.intermediate_folder, 'asl'), 'cbf{:d}'.format(series_idx + 1))

            if series_idx != 0:
                make_nifti(os.path.join(self.intermediate_folder, 'asl'),
                           'cbf{:d}-{:d}'.format(series_idx + 1, series_idx))

    def siemen_calc_yao(self, init_delay, seriesm0):
        scaling_pldcbf = 10
        scaling_wdatt = 100
        aslutil = AnImageASLUtility()
        aslutil.set_vendor('siemens')
        if self.seriesdelay == 5:
            pld_list = [2, 2, 2, 3, 3]
        elif self.seriesdelay == 1:
            pld_list = [10]
        threshold = 0.15
        raw_array = np.zeros((self.row, self.col, len(seriesm0) - 1))
        raw_list = list(seriesm0)
        for i in range(len(raw_list) - 1):
            ds = pydicom.dcmread(seriesm0[raw_list[i + 1]])
            acquisition_idx = ds[0x0020, 0x0012].value - 1
            raw_array[:, :, acquisition_idx] = ds.pixel_array

            # 创建raw_data, 这里可以改成set_raw_data(矩阵）
        raw_array = np.expand_dims(raw_array, axis=2)
        aslutil.set_raw_data(raw_array)

        # 设定参数、内部生成各图片
        aslutil.set_delay_list(init_delay)
        aslutil.set_measure_list(pld_list)

        aslutil.gen_m0()
        aslutil.gen_mask(threshold)
        aslutil.gen_plds()
        if self.seriesdelay == 5:
            aslutil.gen_wd()
            aslutil.gen_att_mcbf_fitting()  # 这个时间长，会显示当前的层序号
            aslutil.gen_cbfs('corrected')  # 这里可以用simple或corrected；simple是将ATT设定为0，跟单延迟一样
            aslutil.gen_acbv()
        elif self.seriesdelay == 1:
            aslutil.gen_cbfs('simple')  # 这里可以用simple或corrected；simple是将ATT设定为0，跟单延迟一样

        # 保存各图象
        if self.debugMode:
            img_plds = aslutil.get_plds()
            for idx in range(self.seriesdelay):
                out_array = img_plds[:, :, :, idx] * scaling_pldcbf
                out_array = np.squeeze(out_array)
                if not os.path.exists(
                        os.path.join(self.intermediate_folder, 'asl', 'pld{:d}'.format(idx + 1))):
                    os.mkdir(os.path.join(self.intermediate_folder, 'asl', 'pld{:d}'.format(idx + 1)))
                SorterUtil.demosaic(ds, out_array, 'pld{:d}'.format(idx + 1), self.intermediate_folder, self.formatpack)

        if self.seriesdelay == 5:
            if self.debugMode == True:
                out_wd = aslutil.get_wd() * scaling_wdatt
                if not os.path.exists(
                        os.path.join(self.intermediate_folder, 'asl', 'wd')):
                    os.mkdir(os.path.join(self.intermediate_folder, 'asl', 'wd'))
                SorterUtil.demosaic(ds, out_wd, 'wd', self.intermediate_folder, self.formatpack)

            out_att = aslutil.get_att() * scaling_wdatt * 10
            if [0x0028, 0x1053] in ds:
                ds[0x0028, 0x1053].value = 0.001
            else:
                ds.add_new([0x0028, 0x1053], 'DS', 0.001)
            if not os.path.exists(
                    os.path.join(self.intermediate_folder, 'asl', 'att')):
                os.mkdir(os.path.join(self.intermediate_folder, 'asl', 'att'))
            SorterUtil.demosaic(ds, out_att, 'att', self.intermediate_folder, self.formatpack)

            out_mcbf = aslutil.get_mcbf() * scaling_pldcbf
            if [0x0028, 0x1053] in ds:
                ds[0x0028, 0x1053].value = 0.1
            else:
                ds.add_new([0x0028, 0x1053], 'DS', 0.1)
            if not os.path.exists(
                    os.path.join(self.intermediate_folder, 'asl', 'mcbf')):
                os.mkdir(os.path.join(self.intermediate_folder, 'asl', 'mcbf'))
            SorterUtil.demosaic(ds, out_mcbf, 'mcbf', self.intermediate_folder, self.formatpack)

            out_acbv = aslutil.get_acbv() * scaling_pldcbf / 60
            if [0x0028, 0x1053] in ds:
                ds[0x0028, 0x1053].value = 0.001
            else:
                ds.add_new([0x0028, 0x1053], 'DS', 0.001)
            if not os.path.exists(
                    os.path.join(self.intermediate_folder, 'asl', 'acbv')):
                os.mkdir(os.path.join(self.intermediate_folder, 'asl', 'acbv'))
            SorterUtil.demosaic(ds, out_acbv, 'acbv', self.intermediate_folder, self.formatpack)

        img_cbfs = aslutil.get_cbfs()
        for idx in range(self.seriesdelay):
            out_cbfs = img_cbfs[:, :, :, idx] * scaling_pldcbf
            if [0x0028, 0x1053] in ds:
                ds[0x0028, 0x1053].value = 0.1
            else:
                ds.add_new([0x0028, 0x1053], 'DS', 0.1)
            if not os.path.exists(
                    os.path.join(self.intermediate_folder, 'asl', 'cbf{:d}'.format(idx + 1))):
                os.mkdir(os.path.join(self.intermediate_folder, 'asl', 'cbf{:d}'.format(idx + 1)))
            SorterUtil.demosaic(ds, out_cbfs, 'cbf{:d}'.format(idx + 1), self.intermediate_folder, self.formatpack)
            if idx != 0:
                out_cbfd = (img_cbfs[:, :, :, idx] - img_cbfs[:, :, :, idx - 1]) * scaling_pldcbf
                # mask=aslutil.get_mask()[:,:,:,idx-1]
                # if np.min(out_cbfd)<0:
                #     modify = (int(abs(np.min(out_cbfd))) // 10 + 1) * 10
                #     out_cbfd[mask==1]+=modify
                #     if [0x0028,0x1052] in ds:
                #         ds[0x0028,0x1052].value=-modify
                #     else:
                #         ds.add_new([0x0028,0x1052], 'DS', -modify)
                if not os.path.exists(
                        os.path.join(self.intermediate_folder, 'asl', 'cbf{:d}-{:d}'.format(idx + 1, idx))):
                    os.mkdir(os.path.join(self.intermediate_folder, 'asl', 'cbf{:d}-{:d}'.format(idx + 1, idx)))
                SorterUtil.demosaic(ds, out_cbfd, 'cbf{:d}-{:d}'.format(idx + 1, idx), self.intermediate_folder,
                                    self.formatpack)

    def parse_pasl_from_archive_folder(self, seriesm0):
        if not os.path.exists(os.path.join(self.intermediate_folder, 'asl', 'mcbf')):
            os.mkdir(os.path.join(self.intermediate_folder, 'asl', 'mcbf'))
        if not os.path.exists(os.path.join(self.intermediate_folder, 'asl', 'att')):
            os.mkdir(os.path.join(self.intermediate_folder, 'asl', 'att'))
        if not os.path.exists(os.path.join(self.intermediate_folder, 'asl', 'm0')):
            os.mkdir(os.path.join(self.intermediate_folder, 'asl', 'm0'))

        seriesm0.pop(0)
        for file_idx in list(seriesm0.values()):
            image_data = pydicom.dcmread(file_idx)
            self.row = image_data[0x0028, 0x0010].value
            self.col = image_data[0x0028, 0x0011].value
            slice_num = image_data[0x0019, 0x100A].value
            sn = math.ceil(np.sqrt(slice_num))
            dmrow = image_data[0x0028, 0x0010].value // sn
            siemensst = image_data[0x0018, 0x0050].value
            siemenssl = image_data[0x0020, 0x1041].value
            siemensipp = image_data[0x0020, 0x0032].value
            siemensiop = [1, 0, 0, 0, 1, 0]
            self.formatpack = [slice_num, sn, dmrow, siemensst, siemenssl, siemensipp, siemensiop]
            if image_data[0x0020, 0x4000].value == 'M0 Image':
                SorterUtil.demosaic(image_data, image_data.pixel_array, 'm0', self.intermediate_folder,
                                    self.formatpack)
            elif image_data[0x0020, 0x4000].value == 'relCBF':
                SorterUtil.demosaic(image_data, image_data.pixel_array, 'mcbf',
                                    self.intermediate_folder,
                                    self.formatpack)
            elif image_data[0x0020, 0x4000].value == 'Bolus Arrival Time (BAT) Image':
                if [0x0028, 0x1053] in image_data:
                    image_data[0x0028, 0x1053].value = 0.001
                else:
                    image_data.add_new([0x0028, 0x1053], 'DS', 0.001)
                SorterUtil.demosaic(image_data, image_data.pixel_array, 'att',
                                    self.intermediate_folder,
                                    self.formatpack)

    def siemens_data_check(self, seriesm0, seriesacbv, seriesatt, seriesmcbf, seriescbf):
        try:
            image_data = pydicom.dcmread(seriesm0[1], stop_before_pixels=True)
        except:
            image_data = pydicom.dcmread([i for i in seriesm0.values()][1], stop_before_pixels=True)
        try:
            slice_num = image_data[0x0019, 0x100A].value
        except:
            raise Exception('slice number not found')
        if self.seriesdelay == 5:
            if len(seriesmcbf) == slice_num + 1:
                for slice_idx in range(slice_num + 1):
                    if slice_idx not in seriesmcbf:
                        raise Exception('missing mcbf slice ' + str(slice_idx))
            else:
                raise Exception('missing mcbf slice')
            if len(seriesatt) == slice_num + 1:
                for slice_idx in range(slice_num + 1):
                    if slice_idx not in seriesatt:
                        raise Exception('missing att slice ' + str(slice_idx))
            else:
                raise Exception('missing att slice')
            if len(seriesacbv) == slice_num + 1:
                for slice_idx in range(slice_num + 1):
                    if slice_idx not in seriesacbv:
                        raise Exception('missing acbv slice ' + str(slice_idx))
            else:
                raise Exception('missing acbv slice')

        if len(seriescbf) != self.seriesdelay:
            raise Exception('number of cbf not equal to series delay')

        for series_idx in range(len(seriescbf)):
            if len(seriescbf[series_idx]) == slice_num + 1:
                for slice_idx in range(slice_num + 1):
                    if slice_idx not in seriescbf[series_idx]:
                        raise Exception('missing cbf slice ' + str(slice_idx))
            else:
                raise Exception('missing slice?')

    def Sorter(self, intermediate_folder, seriesm0, seriesdelay, init_delay=[], seriescbf={}, seriesacbv=[],
               seriesatt=[], seriesmcbf=[], debugMode=False):
        self.intermediate_folder = intermediate_folder
        self.seriesdelay = seriesdelay
        self.debugMode = debugMode
        if not os.path.exists(os.path.join(self.intermediate_folder, 'asl')):
            os.mkdir(os.path.join(self.intermediate_folder, 'asl'))

        if self.seriesdelay == 2:
            self.parse_pasl_from_archive_folder(seriesm0)
        elif (self.seriesdelay == 5 and seriesacbv and seriesatt and seriesmcbf and seriescbf) \
                or (self.seriesdelay == 1 and seriescbf):
            self.siemens_data_check(seriesm0, seriesacbv, seriesatt, seriesmcbf, seriescbf)
            self.parse_m0_from_raw_folder(seriesm0)
            self.parse_cbf_from_archive_folder(seriesmcbf, seriesatt, seriesacbv, seriescbf)
        else:
            self.parse_m0_from_raw_folder(seriesm0)
            self.siemen_calc_yao(init_delay, seriesm0)
